package de.unibi.citec.clf.bonsai.core;

/*
 * #LICENSE%
 * BonSAI [Core|Base]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.ActuatorToConfigure;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.CoreObjectCreationException;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.InitializationException;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.SensorToConfigure;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.SlotToConfigure;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.WorkingMemoryToConfigure;
import de.unibi.citec.clf.bonsai.core.configuration.*;
import de.unibi.citec.clf.bonsai.core.configuration.ConfigurationParser.ParseException;
import de.unibi.citec.clf.bonsai.util.helper.ListClass;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.*;
import java.util.Map.Entry;
import org.apache.log4j.Logger;

/**
 * Central access point to the Bonsai classes. This class provides various
 * methods to create {@link Sensor}s and {@link Actuator}s. The general usage
 * pattern is to first configure the factory via a user-provided configuration
 * file. The default parser is XML-based but can be changed if desired.
 * Afterwards the configured {@link Sensor}s and {@link Actuator}s can be
 * retrieved via the
 * <code>create...</code> methods. A {@link Sensor} itself is not thread-safe if
 * not stated otherwise in the implementing class. The same counts for
 * actuators.
 *
 * This class is implemented as a singleton.
 *
 * @author sebschne
 * @author lkettenb
 * @author nkoester
 * @author jwienke
 * @author lziegler
 */
public class BonsaiManager {

    /**
     * Exception indicating an error in the Bonsai core configuration.
     *
     * @author jwienke
     */
    public class ConfigurationException extends RuntimeException {

        /**
         * Serial version UID.
         */
        private static final long serialVersionUID = -8111248937260206718L;

        /**
         * Constructor.
         */
        public ConfigurationException() {
            super();
        }

        /**
         * Constructor.
         *
         * @param message exception description
         * @param cause cause of the exception
         */
        public ConfigurationException(String message, Throwable cause) {
            super(message, cause);
        }

        /**
         * Constructor.
         *
         * @param message exception description
         */
        public ConfigurationException(String message) {
            super(message);
        }

        /**
         * Constructor.
         *
         * @param cause cause of the exception
         */
        public ConfigurationException(Throwable cause) {
            super(cause);
        }
    }
    private Logger logger = Logger.getLogger(getClass());
    /**
     * The parser instance used for initial configuration.
     *
     * {@link BonsaiManager} does not need a {@link ConfigurationParser}. Use
     * {@link BonsaiManager#configure(java.lang.String, de.unibi.airobots.bonsai.core.configuration.ConfigurationParser)}
     * or
     * {@link BonsaiManager#configure(java.net.URI, de.unibi.airobots.bonsai.core.configuration.ConfigurationParser)}
     * instead.
     */
    @Deprecated
    private ConfigurationParser configurationParser = new XmlConfigurationParser();
    /**
     * A map of configured factories indexed by their class.
     */
    private Map<Class<? extends CoreObjectFactory>, CoreObjectFactory> configuredFactoriesByClass;
    /**
     * Maps every configured working memory by key to its creating factory.
     */
    private Map<String, CoreObjectFactory> memoryKeysToFactories = new HashMap<String, CoreObjectFactory>();
    /**
     * Maps every configured sensor by key to its creating factory.
     */
    private Map<String, CoreObjectFactory> sensorKeysToFactories = new HashMap<String, CoreObjectFactory>();
    /**
     * Maps every configured actuator by key to its creating factory.
     */
    private Map<String, CoreObjectFactory> actuatorKeysToFactories = new HashMap<String, CoreObjectFactory>();
    /**
     * Singleton instance variable.
     */
    private static BonsaiManager instance = new BonsaiManager();

    /**
     * Private constructor for singleton pattern.
     */
    private BonsaiManager() {
    }

    /**
     * Only for testing!!! Don't use this!
     */
    static void testConstructNewInstance() {
        instance = new BonsaiManager();
    }

    /**
     * Returns the singleton instance of this class.
     *
     * @return one and only instance of this class
     */
    public static BonsaiManager getInstance() {
        return instance;
    }

    /**
     * Sets the parser to use for the {@link #configure(String)} call.
     *
     * @param parser new parser to use
     * @deprecated {@link BonsaiManager} does not need a
     * {@link ConfigurationParser}. Use
     * {@link BonsaiManager#configure(java.lang.String, de.unibi.airobots.bonsai.core.configuration.ConfigurationParser)}
     * or
     * {@link BonsaiManager#configure(java.net.URI, de.unibi.airobots.bonsai.core.configuration.ConfigurationParser)}
     * instead.
     */
    @Deprecated
    public void setParser(ConfigurationParser parser) {
        configurationParser = parser;
    }

    /**
     *
     * Returns the parser instance used for initial configuration.
     *
     * @return current parser instance
     * @deprecated Use
     * {@link BonsaiManager#configure(java.lang.String, de.unibi.airobots.bonsai.core.configuration.ConfigurationParser)}
     * or
     * {@link BonsaiManager#configure(java.net.URI, de.unibi.airobots.bonsai.core.configuration.ConfigurationParser)}
     * instead.
     */
    @Deprecated
    public ConfigurationParser getParser() {
        return configurationParser;
    }

    /**
     * Returns the set of configured factory classes after configuration via
     * {@link #configure(URI)}.
     *
     * @return set of configured factory classes
     */
    public Set<Class<? extends CoreObjectFactory>> getConfiguredFactoryClasses() {
        return configuredFactoriesByClass.keySet();
    }

    /**
     * Configures the factory from a given configuration file with the parser
     * set at a call to {@link #setParser(ConfigurationParser)}. Default parser
     * is XML based.
     *
     * @param configurationFilename Path to the local configuration file.
     * @throws IllegalStateException already configured or error while
     * configuring
     * @throws ConfigurationException error configuring the
     * {@link BonsaiManager}. If this is the case, this class is in an
     * unpredictable state and cannot be used afterwards
     *
     * @deprecated Use
     * {@link BonsaiManager#configure(java.lang.String, ConfigurationParser)} or
     * {@link BonsaiManager#configure(java.net.URI, ConfigurationParser)}
     * instead.
     */
    @Deprecated
    public void configure(String configurationFilename)
            throws ConfigurationException {
        try {
            configure(new URI("file://"
                    + configurationFilename.replace(" ", "%20")));
        } catch (URISyntaxException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Configures the factory from a given configuration file with the parser
     * set at a call to {@link #setParser(ConfigurationParser)}. Default parser
     * is XML based.
     *
     * @param configurationFilename Path to the local configuration file.
     * @param parser Parser to parse the given file.
     * @throws ConfigurationException error configuring the
     * {@link BonsaiManager}. If this is the case, this class is in an
     * unpredictable state and cannot be used afterwards
     */
    public void configure(String configurationFilename,
            ConfigurationParser parser) throws ConfigurationException {
        try {
            configure(
                    new URI(configurationFilename.replace(" ", "%20")),
                    parser);
        } catch (URISyntaxException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Configures the factory from a given configuration file with the parser
     * set at a call to {@link #setParser(ConfigurationParser)}. Default parser
     * is XML based.
     *
     * @param configurationFile URI of the configuration file
     * @throws IllegalStateException error while configuring
     * @throws ConfigurationException error configuring the
     * {@link BonsaiManager}. If this is the case, this class is in an
     * unpredictable state and cannot be used afterwards
     *
     * @deprecated Use
     * {@link BonsaiManager#configure(java.lang.String, ConfigurationParser)} or
     * {@link BonsaiManager#configure(java.net.URI, ConfigurationParser)}
     * instead.
     */
    @Deprecated
    public void configure(URI configurationFile) throws IllegalStateException,
            ConfigurationException {
        configure(configurationFile, configurationParser);
    }

    /**
     * Configures the factory from a given configuration file.
     *
     * @param configurationFile URI of the configuration file
     * @param parser Parser to parse the given file.
     * @throws ConfigurationException error configuring the
     * {@link BonsaiManager}. If this is the case, this class is in an
     * unpredictable state and cannot be used afterwards.
     */
    public void configure(URI configurationFile, ConfigurationParser parser)
            throws ConfigurationException {
        configure(configurationFile, parser, null, null, null, null);
    }

    /**
     * Configures the factory from a given configuration file, but only sensors,
     * actuators and memory slots that are consistent with the given maps. These
     * maps may be set null to configure all sensors, actuators or memory slots.
     *
     *
     * @param configurationFile Path to the configuration file
     * @param parser Parser to parse the given file.
     * @param sensorKeySet Map that contains all required sensor keys.
     * @param listSensorKeySet Set that contains all required list sensor keys.
     * @param actuatorKeySet Set that contains all required actuator keys.
     * @param memoryKeySet Set that contains all required working memory keys.
     * @throws ConfigurationException error configuring the
     * {@link BonsaiManager}. If this is the case, this class is in an
     * unpredictable state and cannot be used afterwards.
     */
    public void configure(String configurationFile,
            ConfigurationParser parser,
            Set<String> sensorKeySet,
            Set<String> listSensorKeySet,
            Set<String> actuatorKeySet,
            Set<String> memoryKeySet)
            throws ConfigurationException {
        try {
            configure(new URI(configurationFile.replace(" ", "%20")),
                    parser, sensorKeySet, listSensorKeySet,
                    actuatorKeySet, memoryKeySet);
        } catch (URISyntaxException ex) {
            throw new RuntimeException(ex);
        }
    }

    /**
     * Configures the factory from a given configuration file, but only sensors,
     * actuators and memory slots that are consistent with the given maps. These
     * maps may be set null to configure all sensors, actuators or memory slots.
     *
     *
     * @param configurationFile URI of the configuration file
     * @param parser Parser to parse the given file.
     * @param aSensorKeySet Map that contains all required sensor keys.
     * @param aListSensorKeySet Set that contains all required list sensor keys.
     * @param aActuatorKeySet Set that contains all required actuator keys.
     * @param aMemoryKeySet Set that contains all required working memory keys.
     * @throws ConfigurationException error configuring the
     * {@link BonsaiManager}. If this is the case, this class is in an
     * unpredictable state and cannot be used afterwards.
     */
    public void configure(URI configurationFile,
            ConfigurationParser parser,
            Set<String> aSensorKeySet,
            Set<String> aListSensorKeySet,
            Set<String> aActuatorKeySet,
            Set<String> aMemoryKeySet)
            throws ConfigurationException {
        logger.info("Starting configuration from uri: " + configurationFile);
        Set<String> sensorKeySet = aSensorKeySet == null ? null : new HashSet<String>(aSensorKeySet);
        Set<String> listSensorKeySet = aListSensorKeySet == null ? null : new HashSet<String>(aListSensorKeySet);
        Set<String> actuatorKeySet = aActuatorKeySet == null ? null : new HashSet<String>(aActuatorKeySet);
        Set<String> memoryKeySet = aMemoryKeySet == null ? null : new HashSet<String>(aMemoryKeySet);

        try {
            logger.info("Parsing configuration file");
            parser.parseFile(configurationFile);

            logger.info("Creating factories");
            createFactories(parser.getConfiguredFactories());

            logger.info("Initializing working memories");
            Set<ConfiguredWorkingMemory> requiredWMS =
                    new HashSet<ConfiguredWorkingMemory>();
            if (memoryKeySet != null) {
                Set<ConfiguredWorkingMemory> configuredWMS =
                        parser.getConfiguredWorkingMemories();
                for (ConfiguredWorkingMemory m : configuredWMS) {
                    // Check key
                    if (memoryKeySet.contains(m.getKey())) {
                        requiredWMS.add(m);
                        memoryKeySet.remove(m.getKey());
                    }
                }
            } else {
                requiredWMS = parser.getConfiguredWorkingMemories();
            }
            initializeWorkingMemories(requiredWMS);

            logger.info("Initializing sensors per factory");
            Set<ConfiguredSensor> requiredS =
                    new HashSet<ConfiguredSensor>();
            if (sensorKeySet != null) {
                Set<ConfiguredSensor> configuredS =
                        parser.getConfiguredSensors();
                for (ConfiguredSensor s : configuredS) {
                    // Check key
                    if (sensorKeySet.contains(s.getKey())) {
                        requiredS.add(s);
                        sensorKeySet.remove(s.getKey());
                    }
                    if (listSensorKeySet.contains(s.getKey())) {
                        requiredS.add(s);
                        listSensorKeySet.remove(s.getKey());
                    }
                }
            } else {
                requiredS = parser.getConfiguredSensors();
            }
            initializeSensors(requiredS);

            logger.info("Initializing actuators per factory");
            Set<ConfiguredActuator> requiredA =
                    new HashSet<ConfiguredActuator>();
            if (actuatorKeySet != null) {
                Set<ConfiguredActuator> configuredA =
                        parser.getConfiguredActuators();
                for (ConfiguredActuator a : configuredA) {
                    // Check key
                    if (actuatorKeySet.contains(a.getKey())) {
                        requiredA.add(a);
                        actuatorKeySet.remove(a.getKey());
                    }
                }
            } else {
                requiredA = parser.getConfiguredActuators();
            }
            initializeActuators(requiredA);


            // Required working memory left?
            if (memoryKeySet != null && !memoryKeySet.isEmpty()) {
                String missingM = "";
                for (String key : memoryKeySet) {
                    missingM += "'" + key + "' ";
                }
                throw new ConfigurationException("The following wokring "
                        + " memories have not been configured: " + missingM);
            }
            // Required sensors left?
            if (sensorKeySet != null && !sensorKeySet.isEmpty()) {
                String missingS = "";
                for (String key : sensorKeySet) {
                    missingS += "'" + key + "' ";
                }
                throw new ConfigurationException("The following sensors "
                        + "have not been configured: " + missingS);
            }
            if (listSensorKeySet != null && !listSensorKeySet.isEmpty()) {
                String missingS = "";
                for (String key : listSensorKeySet) {
                    missingS += "'" + key + "' ";
                }
                throw new ConfigurationException("The following sensors "
                        + "have not been configured: " + missingS);
            }
            // Required actuators left?
            if (actuatorKeySet != null && !actuatorKeySet.isEmpty()) {
                String missingA = "";
                for (String key : actuatorKeySet) {
                    missingA += "'" + key + "' ";
                }
                throw new ConfigurationException("The following actuators "
                        + "have not been configured: " + missingA);
            }
        } catch (ParseException e) {
            throw new ConfigurationException(
                    "Error parsing the configuration file \""
                    + configurationFile
                    + "\". Message (ParseException): " + e.getMessage(),
                    e);
        } catch (IOException e) {
            throw new ConfigurationException(
                    "Error parsing the configuration file \""
                    + configurationFile + "\". Message (IOException): "
                    + e.getMessage(), e);
        } catch (IllegalStateException e) {
            throw new ConfigurationException(
                    "Error parsing the configuration file \""
                    + configurationFile
                    + "\". Message (IllegalStateException): "
                    + e.getMessage(), e);
        }
    }

    /**
     * Indicates whether the factory can create a sensor with the unique key
     * <code>key</code> and the data type class
     * <code>dataType</code>.
     *
     * @param key key of the sensor to check
     * @param dataType data type of the sensor with the given key to return
     * @return <code>true</code> if there is a factory that was configured for
     * sensors with the specified key and return data type, else
     * <code>false</code>
     */
    public boolean canCreateSensor(String key, Class<?> dataType) {
        logger.debug("Checking if an sensor of key '" + key
                + "' can be created with interface class " + dataType);

        if (!sensorKeysToFactories.containsKey(key)) {
            logger.debug("No key in map for this sensor.");
            return false;
        }
        if (!sensorKeysToFactories.get(key).canCreateSensor(key,
                dataType)) {
            logger.debug("Factory " + sensorKeysToFactories.get(key)
                    + " cannot create sensor.");
            return false;
        }
        logger.debug("Can create sensor.");
        return true;
    }

    /**
     * Indicates whether the factory can create all given sensors.
     *
     * @param sensors Map of sensor keys and data types.
     * @return <code>true</code> if there is a configured factory for each
     * sensor, else <code>false</code>
     */
    public boolean canCreateSensors(Map<String, Class<?>> sensors) {
        boolean ret = true;
        for (String key : sensors.keySet()) {
            ret &= canCreateSensor(key, sensors.get(key));
        }
        return ret;
    }

    /**
     * Indicates whether the factory can create a sensor with the unique key
     * <code>key</code>, and the list type class
     * <code>listType</code> and the data type class
     * <code>dataType</code>.
     *
     * @param key key of the sensor to check
     * @param listType list type of the sensor with the given key to return
     * @param dataType data type of the list values
     * @return <code>true</code> if there is a factory that was configured for
     * sensor with the specified key and return data type, else
     * <code>false</code>
     */
    public boolean canCreateSensor(String key,
            Class<? extends List<?>> listType, Class<?> dataType) {
        logger.debug("Checking if an sensor of key '" + key
                + "' can be created with interface class " + dataType
                + " and list tpye " + listType);

        if (!sensorKeysToFactories.containsKey(key)) {
            logger.debug("No key in map for this sensor.");
            return false;
        }
        if (!sensorKeysToFactories.get(key).canCreateSensor(key,
                listType, dataType)) {
            logger.debug("Factory " + sensorKeysToFactories.get(key)
                    + " cannot create sensor.");
            return false;
        }
        logger.debug("Can create sensor.");
        return true;
    }

    /**
     * Indicates whether the factory can create all given list sensors.
     *
     * @param sensors Map of sensor keys and {@link ListClass}.
     * @return <code>true</code> if there is a configured factory for each
     * sensor, else <code>false</code>
     */
    public boolean canCreateListSensors(Map<String, ListClass> sensors) {
        boolean ret = true;
        for (String key : sensors.keySet()) {
            ret &= canCreateSensor(key, sensors.get(key).listType,
                    sensors.get(key).dataType);
        }
        return ret;
    }

    /**
     * Tests if an actuator of the desired class can be created. The desired
     * class is interpreted as the minimally fulfilled interface.
     *
     * @param key key of the actuator
     * @param actuatorClass minimal interface an actuator under the given key
     * must implement
     * @return <code>true</code> if such an actuator can be created, else
     * <code>false</code>
     */
    public boolean canCreateActuator(String key,
            Class<? extends Actuator> actuatorClass) {
        logger.debug("Checking if an actuator of key '" + key
                + "' can be created with interface class " + actuatorClass);

        if (!actuatorKeysToFactories.containsKey(key)) {
            logger.debug("No key in map for this actuator.");
            return false;
        }
        if (!actuatorKeysToFactories.get(key).canCreateActuator(key,
                actuatorClass)) {
            logger.debug("Factory " + actuatorKeysToFactories.get(key)
                    + " cannot create actuator.");
            return false;
        }
        logger.debug("Can create actuator.");
        return true;
    }

    /**
     * Indicates whether the factory can create all given actuators.
     *
     * @param actuators Map of actuators keys and data types.
     * @return <code>true</code> if there is a configured factory for each
     * actuator, else <code>false</code>
     */
    public boolean canCreateActuators(Map<String, Class<? extends Actuator>> actuators) {
        boolean ret = true;
        for (String key : actuators.keySet()) {
            ret &= canCreateActuator(key, actuators.get(key));
        }
        return ret;
    }

    /**
     * Tests if an working memory can be created.
     *
     * @param key key of the working memory
     * @return <code>true</code> if such an working memory can be created, else
     * <code>false</code>
     */
    public boolean canCreateWorkingMemory(String key) {
        logger.debug("Checking if an working memory of key '" + key
                + "' can be created.");

        if (!memoryKeysToFactories.containsKey(key)) {
            logger.debug("No key in map for this working memory.");
            return false;
        }
        if (!memoryKeysToFactories.get(key).canCreateWorkingMemory(key)) {
            logger.debug("Factory " + memoryKeysToFactories.get(key)
                    + " cannot create working memory.");
            return false;
        }
        logger.debug("Can create working memory.");
        return true;
    }

    /**
     * Indicates whether the factory can create all given working memories.
     *
     * @param keys Set of working memories keys.
     * @return <code>true</code> if there is a configured factory for each
     * working memory, else <code>false</code>
     */
    public boolean canCreateWorkingMemories(Set<String> keys) {
        boolean ret = true;
        for (String key : keys) {
            ret &= canCreateWorkingMemory(key);
        }
        return ret;
    }

    /**
     * Factory method to create a new sensor for the given data type provided by
     * the resulting sensor. A sensor with this signature can only be created if
     * a call to {@link #canCreateSensor(String, Class)} returned
     * <code>true</code>.
     *
     * @param <T> data type of the sensor to create
     * @param key key under which the sensor is registered in the configuration
     * @param dataType class object representing the data type of the sensor to
     * create
     * @return new instance of the sensor
     * @throws IllegalArgumentException there is no sensor for the given data
     * type with the provided key
     * @throws CoreObjectCreationException error while creating the requested
     * sensor
     */
    public <T> Sensor<T> createSensor(String key, Class<T> dataType)
            throws IllegalArgumentException, CoreObjectCreationException {

        logger.info("Creating a sensor instance with key '" + key
                + "' and dataType '" + dataType + "'");

        if (!sensorKeysToFactories.containsKey(key)) {
            throw new IllegalArgumentException("No sensor with key '" + key
                    + "'");
        }

        CoreObjectFactory factory = sensorKeysToFactories.get(key);
        assert factory != null;
        if (!factory.canCreateSensor(key, dataType)) {
            throw new IllegalArgumentException("Factory " + factory
                    + " cannot create a sensor with key '" + key
                    + "' and dataType '" + dataType + "'.");
        }

        return factory.createSensor(key, dataType);

    }

    /**
     * Factory method to create a new sensor for the given list and data type
     * provided by the resulting sensor. A sensor with this signature can only
     * be created if a call to {@link #canCreateSensor(String, Class, Class)}
     * returned
     * <code>true</code>.
     *
     * @param <S> list type of the sensor to create
     * @param <T> data type of the sensor to create
     * @param key key under which the sensor is registered in the configuration
     * @param listType class object representing the list type of the sensor to
     * create
     * @param dataType class object representing the data type of the sensor to
     * create
     * @return new instance of the sensor
     * @throws IllegalArgumentException there is no sensor for the given data
     * type with the provided key
     * @throws CoreObjectCreationException error while creating the requested
     * sensor
     */
    public <S extends List<T>, T> Sensor<S> createSensor(String key,
            Class<S> listType, Class<T> dataType)
            throws IllegalArgumentException, CoreObjectCreationException {

        logger.info("Creating a sensor instance with key '" + key
                + "', listType '" + listType + "' and dataType '" + dataType
                + "'");

        if (!sensorKeysToFactories.containsKey(key)) {
            throw new IllegalArgumentException("No sensor with key '" + key
                    + "'");
        }

        CoreObjectFactory factory = sensorKeysToFactories.get(key);
        assert factory != null;
        if (!factory.canCreateSensor(key, listType, dataType)) {
            throw new IllegalArgumentException("Factory " + factory
                    + " cannot create a list sensor with key '" + key
                    + "', lsitType '" + listType + "' and dataType '"
                    + dataType + "'.");
        }

        return factory.createSensor(key, listType, dataType);

    }

    /**
     * Factory method to create a new actuator of the provided type.
     *
     * @param <T> type of the actuator to create (minimally implemented
     * interface)
     * @param key key under which the actuator is registered in the
     * configuration
     * @param actuatorClass class object representing the type of the actuator
     * to create
     * @return new instance of the actuator
     * @throws IllegalArgumentException there is no actuator for the given type
     * with the provided key
     * @throws CoreObjectCreationException error while creating the requested
     * actuator
     */
    public <T extends Actuator> T createActuator(String key,
            Class<T> actuatorClass) throws IllegalArgumentException,
            CoreObjectCreationException {

        logger.info("Creating an actuator instance with key '" + key
                + "' and type '" + actuatorClass + "'");

        if (!actuatorKeysToFactories.containsKey(key)) {
            throw new IllegalArgumentException("No actuator with key '" + key
                    + "'");
        }

        CoreObjectFactory factory = actuatorKeysToFactories.get(key);
        assert factory != null;
        if (!factory.canCreateActuator(key, actuatorClass)) {
            throw new IllegalArgumentException("Factory " + factory
                    + " cannot create an actuator with key '" + key
                    + "' and type '" + actuatorClass + "'.");
        }

        return factory.createActuator(key, actuatorClass);

    }

    /**
     * Factory method to create a new working memory.
     *
     * @param key key under which the memory is registered in the configuration
     * @return new instance of the working memory
     * @throws IllegalArgumentException there is no actuator for the given type
     * with the provided key
     * @throws CoreObjectCreationException error while creating the requested
     * actuator
     */
    public <T extends WorkingMemory> T createWorkingMemory(String key)
            throws IllegalArgumentException, CoreObjectCreationException {

        logger.info("Creating an working memory instance with key '" + key
                + "'");

        if (!memoryKeysToFactories.containsKey(key)) {
            throw new IllegalArgumentException("No working memory with key '"
                    + key + "'");
        }

        CoreObjectFactory factory = memoryKeysToFactories.get(key);
        assert factory != null;
        if (!factory.canCreateWorkingMemory(key)) {
            throw new IllegalArgumentException("Factory " + factory
                    + " cannot create an working memory with key '" + key
                    + "'.");
        }

        return factory.createWorkingMemory(key);

    }

    /**
     * Finds the factory with the given class name.
     *
     * @return factory with the given class name or <code>null</code> if not
     * found
     */
    private Class<? extends CoreObjectFactory> findFactory(String className) {
        for (Class<? extends CoreObjectFactory> clazz : configuredFactoriesByClass
                .keySet()) {
            if (clazz.getName().equals(className)) {
                return clazz;
            }
        }
        return null;
    }

    /**
     * Initializes all configured sensors in their appropriate factories found
     * in {@link #configuredFactoriesByClass}. Afterwards the mapping
     * {@link #sensorKeysToFactories} is populated.
     *
     * @param configuredSensors sensors to initialize
     * @throws ConfigurationExceptionerror initialization error
     */
    private void initializeSensors(Set<ConfiguredSensor> configuredSensors)
            throws ConfigurationException {
        sensorKeysToFactories = new HashMap<String, CoreObjectFactory>();

        // first, map every configured sensor to the factory that creates it
        Map<CoreObjectFactory, Set<SensorToConfigure>> sensorToConfigureByFactory = mapConfiguredCoreObjectToFactories(
                configuredSensors,
                new CoreObjectMapper<ConfiguredSensor, SensorToConfigure>() {
            @Override
            public SensorToConfigure map(ConfiguredSensor t)
                    throws ConfigurationException {
                return mapSensor(t);
            }
        });

        // then, initialize the sensors of every factory
        for (Entry<CoreObjectFactory, Set<SensorToConfigure>> factoryEntry : sensorToConfigureByFactory
                .entrySet()) {
            try {
                logger.debug("Factory " + factoryEntry.getKey() + ": "
                        + factoryEntry.getValue());
                factoryEntry.getKey().configureSensors(factoryEntry.getValue());
            } catch (IllegalArgumentException e) {
                String message = "Error configuring a sensor for factory "
                        + factoryEntry.getKey().getClass().getSimpleName();
                logger.error(message);
                throw new ConfigurationException(message, e);
            } catch (CoreObjectCreationException e) {
                throw new ConfigurationException(e);
            }
        }

        // if the initialization was successful, populate the map of sensor
        // keys to factory objects
        logger.info("Building map of initialized sensors");
        for (Entry<CoreObjectFactory, Set<SensorToConfigure>> factoryEntry : sensorToConfigureByFactory
                .entrySet()) {

            CoreObjectFactory factory = factoryEntry.getKey();
            logger.debug("   factory = " + factory);

            for (SensorToConfigure sensor : factoryEntry.getValue()) {

                logger.debug("      sensor = " + sensor);

                if (sensorKeysToFactories.containsKey(sensor.getKey())) {
                    throw new ConfigurationException(
                            "There is already a sensor configured with key '"
                            + sensor.getKey() + "'. "
                            + "Cannot initialize another sensor"
                            + " with this key: " + sensor);
                }

                sensorKeysToFactories.put(sensor.getKey(), factory);

            }

        }

    }

    /**
     * Initializes all configured actuators in their appropriate factories found
     * in {@link #configuredFactoriesByClass}. Afterwards the mapping
     * {@link #actuatorKeysToFactories} is populated.
     *
     * @param configuredActuators actuators to initialize
     * @throws ConfigurationExceptionerror initialization error
     */
    private void initializeActuators(Set<ConfiguredActuator> configuredActuators) {
        actuatorKeysToFactories = new HashMap<String, CoreObjectFactory>();
        // map requested actuators to the factories that shall create them
        Map<CoreObjectFactory, Set<ActuatorToConfigure>> actuatorsToConfigureByFactory = mapConfiguredCoreObjectToFactories(
                configuredActuators,
                new CoreObjectMapper<ConfiguredActuator, ActuatorToConfigure>() {
            @Override
            public ActuatorToConfigure map(ConfiguredActuator t)
                    throws ConfigurationException {
                return mapActuator(t);
            }
        });

        // initialize actuators in each factory
        for (Entry<CoreObjectFactory, Set<ActuatorToConfigure>> factoryEntry : actuatorsToConfigureByFactory
                .entrySet()) {
            try {
                logger.debug("Factory " + factoryEntry.getKey() + ": "
                        + factoryEntry.getValue());
                factoryEntry.getKey().configureActuators(
                        factoryEntry.getValue());
            } catch (IllegalArgumentException e) {
                String message = "Error configuring an actuator for factory "
                        + factoryEntry.getKey().getClass().getSimpleName();
                logger.error(message);
                throw new ConfigurationException(message, e);
            } catch (CoreObjectCreationException e) {
                throw new ConfigurationException(e);
            }
        }

        // build a mapping of actuator keys to creating factories
        logger.info("Building map of initialized actuators");
        for (Entry<CoreObjectFactory, Set<ActuatorToConfigure>> factoryEntry : actuatorsToConfigureByFactory
                .entrySet()) {

            CoreObjectFactory factory = factoryEntry.getKey();
            logger.debug("   factory = " + factory);

            for (ActuatorToConfigure actuator : factoryEntry.getValue()) {

                logger.debug("      actuator = " + actuator);

                if (actuatorKeysToFactories.containsKey(actuator.getKey())) {
                    throw new ConfigurationException(
                            "There is already an actuator configured with key '"
                            + actuator.getKey() + "'. "
                            + "Cannot initialize another actuator "
                            + "with this key: " + actuator);
                }

                actuatorKeysToFactories.put(actuator.getKey(), factory);

            }

        }

    }

    /**
     * Initializes all configured working memories in their appropriate
     * factories found in {@link #configuredFactoriesByClass}. Afterwards the
     * mapping {@link #memoryKeysToFactories} is populated.
     *
     * @param configuredMemories working memories to initialize
     * @throws ConfigurationExceptionerror initialization error
     */
    private void initializeWorkingMemories(
            Set<ConfiguredWorkingMemory> configuredMemories) {

        memoryKeysToFactories = new HashMap<String, CoreObjectFactory>();
        // map requested actuators to the factories that shall create them
        Map<CoreObjectFactory, Set<WorkingMemoryToConfigure>> memoriesToConfigureByFactory = mapConfiguredCoreObjectToFactories(
                configuredMemories,
                new CoreObjectMapper<ConfiguredWorkingMemory, WorkingMemoryToConfigure>() {
            @Override
            public WorkingMemoryToConfigure map(
                    ConfiguredWorkingMemory t)
                    throws ConfigurationException {
                return mapWorkingMemory(t);
            }
        });

        // initialize memories in each factory
        for (Entry<CoreObjectFactory, Set<WorkingMemoryToConfigure>> factoryEntry : memoriesToConfigureByFactory
                .entrySet()) {
            try {
                logger.debug("Factory " + factoryEntry.getKey() + ": "
                        + factoryEntry.getValue());
                factoryEntry.getKey().configureWorkingMemories(
                        factoryEntry.getValue());
            } catch (IllegalArgumentException e) {
                String message = "Error configuring a working memory for factory "
                        + factoryEntry.getKey().getClass().getSimpleName();
                logger.error(message);
                logger.error(e.getMessage());
                throw new ConfigurationException(message, e);
            } catch (CoreObjectCreationException e) {
                throw new ConfigurationException(e);
            }
        }

        // build a mapping of slot keys to creating factories
        logger.info("Building map of initialized working memorys");
        for (Entry<CoreObjectFactory, Set<WorkingMemoryToConfigure>> factoryEntry : memoriesToConfigureByFactory
                .entrySet()) {

            CoreObjectFactory factory = factoryEntry.getKey();
            logger.debug("   factory = " + factory);

            for (WorkingMemoryToConfigure memory : factoryEntry.getValue()) {

                logger.debug("      working memory = " + memory);

                if (memoryKeysToFactories.containsKey(memory.getKey())) {
                    throw new ConfigurationException(
                            "There is already an working memory configured with key '"
                            + memory.getKey()
                            + "'. "
                            + "Cannot initialize another working memory "
                            + "with this key: " + memory);
                }

                memoryKeysToFactories.put(memory.getKey(), factory);

            }

        }

    }

    /**
     * Maps {@link ConfiguredActuator}s to {@link ActuatorToConfigure}s with
     * appropriate class checking.
     *
     * @param confAct actuator to map
     * @return mapped actuator
     * @throws ConfigurationException error while mapping, especially class
     * finding errors
     */
    @SuppressWarnings("unchecked")
    private ActuatorToConfigure mapActuator(ConfiguredActuator confAct)
            throws ConfigurationException {

        ActuatorToConfigure actuator = new ActuatorToConfigure();

        actuator.setKey(confAct.getKey());
        actuator.setActuatorOptions(confAct.getOptions());

        try {

            Class<?> possibleInterfaceClass = Class.forName(confAct
                    .getActuatorClassName());
            if (!Actuator.class.isAssignableFrom(possibleInterfaceClass)) {
                throw new ConfigurationException("Class "
                        + possibleInterfaceClass
                        + " does not implement interface " + Actuator.class);
            }
            actuator.setInterfaceClass((Class<? extends Actuator>) possibleInterfaceClass);

            Class<?> possibleActuatorClass = Class.forName(confAct
                    .getActuatorClassName());
            if (!Actuator.class.isAssignableFrom(possibleActuatorClass)) {
                throw new ConfigurationException("Class "
                        + possibleActuatorClass
                        + " does not implement interface " + Actuator.class);
            }
            if (!actuator.getInterfaceClass().isAssignableFrom(
                    possibleActuatorClass)) {
                throw new ConfigurationException("Class "
                        + possibleActuatorClass + " does not implement its "
                        + "declared actuator interface "
                        + actuator.getInterfaceClass());
            }
            actuator.setActuatorClass((Class<? extends Actuator>) possibleActuatorClass);

        } catch (ExceptionInInitializerError e) {
            throw new ConfigurationException(e);
        } catch (LinkageError e) {
            throw new ConfigurationException(e);
        } catch (ClassNotFoundException e) {
            throw new ConfigurationException(e);
        }

        return actuator;

    }

    /**
     * An interface describing a method to map one data type to another. This is
     * used to map configuration objects from {@link BonsaiManager} to instances
     * of {@link ConfiguredXcfCoreObject}.
     *
     * @author jwienke
     * @param <T> source data type for the mapping
     * @param <U> target data type for the mapping
     */
    private interface CoreObjectMapper<T, U> {

        U map(T t) throws ConfigurationException;
    }

    /**
     * Maps all configured sensors given to the factory instances that create
     * the sensor.
     *
     * @param <T> type of the core objects to map
     * @param configuredObjects sensors to map
     * @param mapper mapper from <code>T</code> to <code>U</code>
     * @return map that maps every given {@link ConfiguredSensor} to the factory
     * object that can create it
     * @throws ConfigurationException on of the sensors cannot be mapped to a
     * factory instance
     */
    private <T extends ConfiguredCoreObject, U> Map<CoreObjectFactory, Set<U>> mapConfiguredCoreObjectToFactories(
            Set<T> configuredObjects, CoreObjectMapper<T, U> mapper)
            throws ConfigurationException {

        logger.debug("Mapping ConfiguredObject to factories. Got "
                + configuredObjects.size() + " object(s) to map.");

        Map<CoreObjectFactory, Set<U>> coreObjectByFactory = new HashMap<CoreObjectFactory, Set<U>>();

        for (T coreObject : configuredObjects) {

            Class<? extends CoreObjectFactory> factoryClass = findFactory(coreObject
                    .getFactoryClassName());
            if (factoryClass == null) {
                throw new ConfigurationException(
                        "Cannot setup sensor with key '" + coreObject.getKey()
                        + "' because there is no factory set up "
                        + "with class '"
                        + coreObject.getFactoryClassName() + "'");
            }

            CoreObjectFactory factory = configuredFactoriesByClass
                    .get(factoryClass);
            assert factory != null : "We could find a factory with class "
                    + factoryClass + " so it must be contained in"
                    + "the map of configured factories.";
            if (!coreObjectByFactory.containsKey(factory)) {
                coreObjectByFactory.put(factory, new HashSet<U>());
            }
            logger.debug(coreObject + " belongs to " + factory);
            boolean notReplaced = coreObjectByFactory.get(factory).add(
                    mapper.map(coreObject));
            assert notReplaced;

        }

        if (logger.isDebugEnabled()) {
            for (Entry<CoreObjectFactory, Set<U>> entry : coreObjectByFactory
                    .entrySet()) {
                StringBuilder builder = new StringBuilder();
                builder.append("Mapping of objects to factories:\n");
                builder.append("   " + entry.getKey() + ":\n");
                for (U s : entry.getValue()) {
                    builder.append("      " + s + "\n");
                }
                logger.debug(builder.toString());
            }
        }

        return coreObjectByFactory;

    }

    /**
     * Maps {@link ConfiguredSensor}s to {@link ActuatorToConfigure}s with
     * appropriate class checking.
     *
     * @param confSensor sensor to map
     * @return mapped sensor
     * @throws ConfigurationException error while mapping, especially class
     * finding errors
     */
    @SuppressWarnings("unchecked")
    private SensorToConfigure mapSensor(ConfiguredSensor confSensor)
            throws ConfigurationException {

        SensorToConfigure sensor = new SensorToConfigure();

        sensor.setKey(confSensor.getKey());
        sensor.setSensorOptions(confSensor.getOptions());

        try {

            sensor.setDataTypeClass(Class.forName(confSensor
                    .getDataTypeClassName()));

            if (confSensor.isListSensor()) {

                Class<?> possibleListClass = Class.forName(confSensor
                        .getListTypeClassName());
                if (!List.class.isAssignableFrom(possibleListClass)) {
                    throw new ConfigurationException("Class "
                            + possibleListClass
                            + " does not implement interface " + List.class);
                }
                sensor.setListTypeClass((Class<? extends List<?>>) possibleListClass);
            }

            Class<?> possibleSensorClass = Class.forName(confSensor
                    .getSensorClassName());
            if (!Sensor.class.isAssignableFrom(possibleSensorClass)) {
                throw new ConfigurationException("Class " + possibleSensorClass
                        + " does not implement interface " + Sensor.class);
            }
            sensor.setSensorClass((Class<? extends Sensor<?>>) possibleSensorClass);

        } catch (ExceptionInInitializerError e) {
            throw new ConfigurationException(e);
        } catch (LinkageError e) {
            throw new ConfigurationException(e);
        } catch (ClassNotFoundException e) {
            throw new ConfigurationException(e);
        }

        return sensor;

    }

    /**
     * Maps {@link ConfiguredWorkingMemory}s to {@link WorkingMemoryToConfigure}
     * s with appropriate class checking.
     *
     * @param confMemory memory to map
     * @return mapped memory
     * @throws ConfigurationException error while mapping, especially class
     * finding errors
     */
    @SuppressWarnings("unchecked")
    private WorkingMemoryToConfigure mapWorkingMemory(
            ConfiguredWorkingMemory confMemory) throws ConfigurationException {

        WorkingMemoryToConfigure memory = new WorkingMemoryToConfigure();

        memory.setKey(confMemory.getKey());
        memory.setMemoryOptions(confMemory.getOptions());

        try {

            Set<SlotToConfigure> slots = new HashSet<SlotToConfigure>();
            for (ConfiguredMemorySlot confSlot : confMemory.getMemorySlots()) {
                SlotToConfigure slot = new SlotToConfigure();
                slot.setDataTypeClass(Class.forName(confSlot
                        .getDataTypeClassName()));
                Class<?> possibleSlotClass = Class.forName(confSlot
                        .getSlotClassName());
                if (!MemorySlot.class.isAssignableFrom(possibleSlotClass)) {
                    throw new ConfigurationException("Class "
                            + possibleSlotClass
                            + " does not implement interface "
                            + MemorySlot.class);
                }
                slot.setSlotClass((Class<? extends MemorySlot<?>>) possibleSlotClass);
                slots.add(slot);
            }
            memory.setSlots(slots);

            Class<?> possibleMemoryClass = Class.forName(confMemory
                    .getMemoryClassName());
            if (!WorkingMemory.class.isAssignableFrom(possibleMemoryClass)) {
                throw new ConfigurationException("Class " + possibleMemoryClass
                        + " does not implement interface "
                        + WorkingMemory.class);
            }
            memory.setMemoryClass((Class<? extends WorkingMemory>) possibleMemoryClass);

        } catch (ExceptionInInitializerError e) {
            throw new ConfigurationException(e);
        } catch (LinkageError e) {
            throw new ConfigurationException(e);
        } catch (ClassNotFoundException e) {
            throw new ConfigurationException(e);
        }

        return memory;

    }

    /**
     * Creates the collection of user configured factories.
     *
     * @param requestedFactories set of factory settings found in the
     * configuration file.
     * @throws ConfigurationException error configuring a factory
     */
    private void createFactories(
            Set<ConfiguredCoreObjectFactory> requestedFactories)
            throws ConfigurationException {

        logger.debug("Creating factories: " + requestedFactories);

        if (requestedFactories.size() < 1) {
            throw new ConfigurationException(
                    "No factories requested by the user. "
                    + "This does not make sense, because then "
                    + "no sensor or actuator can be created.");
        }

        for (ConfiguredCoreObjectFactory requestedFactory : requestedFactories) {
            createFactory(requestedFactory);
        }

    }

    /**
     * Instantiates a configured factory and adds it to
     * {@link #configuredFactoriesByClass}.
     *
     * @param requestedFactory factory to configure
     * @throws ConfigurationException error configuring the factory
     */
    private void createFactory(ConfiguredCoreObjectFactory requestedFactory)
            throws ConfigurationException {
        configuredFactoriesByClass = new HashMap<Class<? extends CoreObjectFactory>, CoreObjectFactory>();
        try {

            // first check that a class with the desired name exists and can be
            // instantiated
            Class<?> rawFactoryClass = Class.forName(requestedFactory
                    .getFactoryClassName());
            Object rawInstance = rawFactoryClass.newInstance();
            if (!(rawInstance instanceof CoreObjectFactory)) {
                throw new ConfigurationException("Class '"
                        + requestedFactory.getFactoryClassName()
                        + "' is not an instance of "
                        + CoreObjectFactory.class.getSimpleName());
            }
            CoreObjectFactory newFactory = (CoreObjectFactory) rawInstance;

            Class<? extends CoreObjectFactory> factoryClass = newFactory
                    .getClass();

            // configure the new factory
            try {
                newFactory.initialize(requestedFactory.getFactroyOptions());
            } catch (IllegalArgumentException e) {
                throw new ConfigurationException(
                        "Error configuring factory of class " + factoryClass.getSimpleName()
                        + " with the user supplied options: "
                        + requestedFactory.getFactroyOptions(), e);
            } catch (InitializationException e) {
                throw new ConfigurationException(
                        "Error initializing factory of class " + factoryClass.getSimpleName()
                        + ".", e);
            }

            // add it to the list of configured factories
            configuredFactoriesByClass.put(factoryClass, newFactory);

        } catch (ClassNotFoundException e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(
                    "Could not find a class with name '"
                    + requestedFactory.getFactoryClassName() + "'", e);
        } catch (ExceptionInInitializerError e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(e);
        } catch (LinkageError e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(e);
        } catch (InstantiationException e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(
                    "Error creating a new instance of class '"
                    + requestedFactory.getFactoryClassName()
                    + "'. Is it a top level member?", e);
        } catch (IllegalAccessException e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(
                    "Error creating a new instance of class '"
                    + requestedFactory.getFactoryClassName()
                    + "'. Does it have a default constructor?", e);
        } catch (SecurityException e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(
                    "Error creating a new instance of class '"
                    + requestedFactory.getFactoryClassName() + "'", e);
        }

    }
}
