/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Robin Schiewer, Sebastian 
 * Meyer zu Borgsen, Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Tristan 
 * Walter, Viktor Losing, Viktor Richter
 */
package de.unibi.citec.clf.bonsai.xcf;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.CommunicationException;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.CoreObjectCreationException;
import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.InitializationException;
import de.unibi.citec.clf.bonsai.core.MemorySlot;
import de.unibi.citec.clf.bonsai.core.WorkingMemory;
import de.unibi.citec.clf.bonsai.xcf.configuration.ConfiguredXcfAmWorkingMemorySlot;
import java.util.Map;
import java.util.Set;
import net.sf.xcf.ActiveMemory;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.util.QueueClosedException;
import net.sf.xcf.util.SynchronizedQueue;
import net.sf.xcf.xml.XPath;
import org.apache.log4j.Logger;

public class AmWorkingMemory implements WorkingMemory,
        ActiveMemoryReconnectable {

    /**
     * Option for publisher based sensors defining the memory name.
     */
    public static final String OPTION_MEMORY_NAME = "memoryName";
    /**
     * Instance of the active memory that is used.
     */
    private ActiveMemory activeMemory = null;
    /**
     * Name of the active memory.
     */
    private String memoryName = "";
    /**
     * Lock for synchronized memory access.
     */
    private final Object memoryLock = new Object();
    /**
     * The log.
     */
    private Logger logger = Logger.getLogger(this.getClass());
    private Set<ConfiguredXcfAmWorkingMemorySlot> slots = null;

    /**
     * Constructor.
     *
     * @param memoryName Name of the active memory to work on.
     */
    public AmWorkingMemory(String memoryName,
            Set<ConfiguredXcfAmWorkingMemorySlot> slots) {
        this.memoryName = memoryName;
        this.slots = slots;
    }

    @Override
    public void initialize(Map<String, String> options)
            throws IllegalArgumentException, InitializationException {

        if (!options.containsKey(OPTION_MEMORY_NAME)) {
            throw new IllegalArgumentException(
                    "Memory name missing in arguments. Add option \""
                    + OPTION_MEMORY_NAME
                    + "\" to bonsai configuration.");
        }

        memoryName = options.get(OPTION_MEMORY_NAME);
    }

    @Override
    public <T> MemorySlot<T> getSlot(String xpathPrefix, Class<T> dataType)
            throws CommunicationException, IllegalArgumentException,
            CoreObjectCreationException {

        // find suitable slot
        for (ConfiguredXcfAmWorkingMemorySlot confSlot : slots) {
            if (confSlot.isSuitableFor(dataType)) {
                logger.debug("Creating \""
                        + confSlot.getSlotClass().getSimpleName()
                        + "\" for slot: " + xpathPrefix);
                AmWorkingMemorySlot<T> slot = confSlot.createInstance();
                slot.initialize(this, xpathPrefix, dataType);

                return slot;
            }
        }

        throw new CoreObjectCreationException(
                "Can not create MemorySlot for data type: "
                + dataType.getSimpleName() + ". No applicable slot "
                + "implementation found in configuration.");
    }

    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }

    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {
        synchronized (memoryLock) {
            this.activeMemory = activeMemory;
        }
    }

    void memorize(XOPData data) throws CommunicationException {
        try {
            logger.trace("Memorizing");
            logger.trace("XOP: " + data.getDocumentAsText());
            activeMemory.insert(data);
            logger.trace("done");
        } catch (MemoryException e) {
            throw new CommunicationException(
                    "Error sending data to Active Memory", e);
        }
    }

    void forget(String slot) throws CommunicationException {
        try {
            activeMemory.remove(new XPath(slot));
        } catch (MemoryException e) {
            throw new CommunicationException(
                    "Error removing data from Active Memory", e);
        }
    }

    XOPData recall(String slot) throws CommunicationException {
        try {
            SynchronizedQueue<XOPData> data = activeMemory.query(
                    new XPath(slot), false);

            // get newest data
            XOPData d = null;
            do {
                try {
                    d = data.next();
                } catch (QueueClosedException e) {
                    logger.warn("Queue was closed while accessing");
                    return null;
                }

            } while (data.hasNext());
            return d;
        } catch (MemoryException e) {
            throw new CommunicationException(
                    "Error receiving data from Active Memory", e);
        }
    }
}
