package de.unibi.citec.clf.bonsai.xcf;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import net.sf.xcf.util.InterruptedException;
import net.sf.xcf.util.QueueClosedException;
import net.sf.xcf.util.SynchronizedQueue;

/**
 * Creates a synchronized bound queue with a limited capacity. Before pushing an
 * object into the queue it removes elements so that the size of the queue is
 * always within the capacity.
 * 
 * @author marc
 * @author jwienke
 * @param <T>
 *            type of elements contained in this queue
 */
public abstract class BoundSynchronizedQueue<T> extends SynchronizedQueue<T> {

	private int capacity;
	private T cache;

	/**
	 * Returns the maximum capacity of this queue.
	 * 
	 * @return the capacity
	 */
	public int getCapacity() {
		return capacity;
	}

	/**
	 * Sets the maximum capacity of this queue.
	 * 
	 * @param capacity
	 *            the capacity to set
	 */
	public void setCapacity(int capacity) {
		this.capacity = capacity;
		// TODO what if this queue contains more entries than the new capacity?
	}

	/**
	 * Create queue with a given capacity.
	 * 
	 * @param capacity
	 *            the initial capacity.
	 */
	public BoundSynchronizedQueue(int capacity) {
		super();
		this.capacity = capacity;
	}

	/**
	 * Create queue with a capacity of 1. This implements a synchronized buffer.
	 */
	public BoundSynchronizedQueue() {
		super();
		this.capacity = 1;
	}

	@Override
	public synchronized void push(T element) {
		if (super.isClosed()) {
			throw new QueueClosedException("Queue is closed!");
		}
		// discard all older elements

		while (super.getSize() >= capacity) {
			super.pop();
		}
		super.push(element);
	}

	/**
	 * Retrieves the head of this queue. This method differs from the pop method
	 * in that it waits for a new element for a given timeout if the queue is
	 * empty. When the waiting thread is interrupted an
	 * {@link InterruptedException} will be thrown. If the timeout is reached
	 * the last known position is returned.
	 * 
	 * @param timeout
	 *            the maximum time waiting for a new element in an empty queue
	 *            in milliseconds
	 * @return the head of this queue
	 * @throws InterruptedException
	 *             if thread is interrupted
	 */
	public synchronized T nextCached(long timeout) throws InterruptedException {
		T data = next(timeout);
		if (data == null) {
			return cache;
		} else {
			cache = data;
			return data;
		}
	}
}
