package de.unibi.citec.clf.bonsai.xcf;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Map.Entry;

import net.sf.xcf.InitializeException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.XcfManager;
import net.sf.xcf.naming.NameNotFoundException;

import org.apache.log4j.Logger;

import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable.ErrorHandlingPolicy;

/**
 * This class manages {@link RemoteServerReconnectable}s and ensures that their
 * remote server is reset if this server died and is accessible again.
 * 
 * @author jwienke
 */
public class RemoteServerReconnectableManager extends
        ReconnectableManager<RemoteServerReconnectable> {

    private Logger logger = Logger.getLogger(getClass());

    /**
     * A mutex object for {@link #remoteServerWrapperByName}.
     */
    private final Object remoteServerLock = new Object();

    /**
     * The {@link XcfManager} to use.
     */
    private XcfManager xcfManager = null;

    /**
     * The map of remote servers currently connected to or if value is
     * <code>null</code> a connection is wanted but isn't possible right now.
     * 
     * Always access this object by acquiring {@link #remoteServerLock} before.
     */
    private Map<String, RemoteServer> remoteServersByName =
            new HashMap<String, RemoteServer>();

    /**
     * Map of managed {@link ConnectionErrorResistantRemoteServer}s by their
     * server name.
     * 
     * Always synchronize on {@link #remoteServerLock} before using this map.
     */
    private Map<String, Set<ConnectionErrorResistantRemoteServer>> decoratorsByServerName =
            new HashMap<String, Set<ConnectionErrorResistantRemoteServer>>();

    /**
     * Map of managed {@link RemoteServerReconnectable}s that want to be
     * directly informed on failing remote servers without using the decorator.
     * 
     * Always synchronize on {@link #remoteServerLock} before using this map.
     */
    private Map<String, Set<RemoteServerReconnectable>> directNotifiers =
            new HashMap<String, Set<RemoteServerReconnectable>>();

    /**
     * Timer used to schedule the checking for subscribers.
     */
    private Timer checkTimer = new Timer("remote server check timer");

    /**
     * The actual task executed by {@link #checkTimer} to check for remote
     * servers.
     */
    private RemoteServerChecker remoteServerCheckTask =
            new RemoteServerChecker();

    /**
     * Just a utility for inner classes to avoid <code>this</code> confusion.
     */
    private final RemoteServerReconnectableManager instance = this;

    /**
     * Thread that checks all remote servers for being alive.
     * 
     * @author jwienke
     */
    private class RemoteServerChecker extends TimerTask {

        /**
         * {@inheritDoc}
         */
        @Override
        public void run() {

            logger.debug("Checking remote servers starts...");

            // May reduce scope of this lock if this becomes a performance
            // issue. This may be the case if new reconnectables need to be
            // managed from time to time while checking already runs.
            synchronized (remoteServerLock) {

                // iterate over all managed remote servers
                for (Entry<String, RemoteServer> remoteServerEntry : remoteServersByName
                        .entrySet()) {

                    String serverName = remoteServerEntry.getKey();
                    RemoteServer lastServer = remoteServerEntry.getValue();

                    checkRemoteServer(serverName, lastServer);

                }

            }

            logger.debug("Checking remote servers finished");

        }

        /**
         * Does the check for one single remote server. Assumes that
         * <code>remoteServerLock</code> is acquired.
         * 
         * @param serverName
         *            the name of the server to check
         * @param lastServer
         *            last server instance or <code>null</code> if this server
         *            was already disconnected
         */
        private void checkRemoteServer(String serverName,
                RemoteServer lastServer) {

            try {

                // try to connect to this subscriber
                RemoteServer checkRemoteServer =
                        xcfManager.createRemoteServer(serverName);

                logger.debug("RemoteServer '" + serverName
                        + "' seems to be alive.");

                // if connection succeeded and there was no connection
                // before this attempt, use the new server and notify
                // decorators or reconnectables that want to be informed
                // directly
                if (lastServer == null) {

                    logger.info("RemoteServer '" + serverName
                            + "' was successfully (re-)connected. "
                            + "Informing decorators about this.");

                    remoteServersByName.put(serverName, checkRemoteServer);

                    // TODO are remote servers reentrant? Otherwise we need to
                    // create a new instance for every interested client

                    // inform decoartors
                    if (decoratorsByServerName.containsKey(serverName)) {
                        logger.debug("Informing decorators about new server");
                        for (ConnectionErrorResistantRemoteServer decorator : decoratorsByServerName
                                .get(serverName)) {
                            logger.debug("Informing decorator " + decorator
                                    + " about new server");
                            decorator.resetRemoteServer(checkRemoteServer);
                        }
                    }

                    // inform direct reconnectables
                    if (directNotifiers.containsKey(serverName)) {
                        logger.debug("Informing direct reconnectables "
                                + "about new server");
                        for (RemoteServerReconnectable reconnectable : directNotifiers
                                .get(serverName)) {
                            logger.debug("Resetting remote server "
                                    + "in reconnectable " + reconnectable);
                            reconnectable
                                    .resetRemoteServer(new ConnectionErrorNotifyingRemoteServer(
                                            checkRemoteServer, instance));
                        }
                    }

                } else {
                    checkRemoteServer.deactivate();
                }

                // CHECKSTYLE:OFF catch every possible error to not
                // break the checking thread
            } catch (Throwable e) {
                // CHECKSTYLE:ON

                // connection to subscriber failed, mark it as
                // unconnected

                logger.warn("RemoteServer with name '" + serverName + "'"
                        + " could not be connected "
                        + "or reinitializing in of the reconnectables. "
                        + "Marking this remote server as disconnected.", e);

                markRemoteServerAsDisconnected(serverName);

            }

        }

    }

    /**
     * Constructor using default check interval
     * {@link ReconnectableManager#DEFAULT_CHECK_INTERVAL_IN_MS}.
     * 
     * @param xcfManager
     *            xcf manager to use
     */
    public RemoteServerReconnectableManager(XcfManager xcfManager) {
        this(xcfManager, DEFAULT_CHECK_INTERVAL_IN_MS);
    }

    /**
     * Constructor.
     * 
     * @param xcfManager
     *            xcf manager to use for xcf operations
     * @param checkIntervalMs
     *            interval for the checking task
     */
    public RemoteServerReconnectableManager(XcfManager xcfManager,
            long checkIntervalMs) {
        this.xcfManager = xcfManager;

        checkTimer.scheduleAtFixedRate(remoteServerCheckTask, checkIntervalMs,
                checkIntervalMs);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void initializeXcfTypes(Set<String> remoteServerNames)
            throws XcfTypeInitializationException {

        logger.info("Initializing remote servers: " + remoteServerNames);

        synchronized (remoteServerLock) {

            for (String remoteServerName : remoteServerNames) {
                if (!remoteServerName.equals("")) {
                    if (!remoteServersByName.containsKey(remoteServerName)) {
                        createRemoteServer(remoteServerName);
                    }
                }
            }

        }

    }

    /**
     * Tries to create a remote server with the specified name. If the creation
     * fails and {@link #errorOnRemoteServerInitialization} is <code>true</code>
     * , an exception will be thrown. Otherwise a <code>null</code> entry will
     * be added to {@link #remoteServerWrapperByName} requesting the postponed
     * initialization. Also, successfully created {@link RemoteServer} instances
     * are added to that map.
     * 
     * @param remoteServerName
     *            name of the remote server to create
     * @throws XcfTypeInitializationException
     *             thrown if a connection to a remote server with the specified
     *             name is not possible and
     *             {@link #errorOnRemoteServerInitialization} is
     *             <code>true</code>
     */
    private void createRemoteServer(String remoteServerName)
            throws XcfTypeInitializationException {

        synchronized (remoteServerLock) {

            try {

                RemoteServer remoteServer =
                        xcfManager.createRemoteServer(remoteServerName);
                remoteServersByName.put(remoteServerName, remoteServer);

            } catch (NameNotFoundException e) {

                if (isErrorOnXcfTypeInitialization()) {
                    throw new XcfTypeInitializationException(
                            "No server with name '" + remoteServerName
                                    + "' found.", e);
                } else {
                    logger.warn("Right now no remote server for name '"
                            + remoteServerName + "' could be created. "
                            + "Initialization will be delayed.", e);
                    remoteServersByName.put(remoteServerName, null);
                }

            } catch (InitializeException e) {

                if (isErrorOnXcfTypeInitialization()) {
                    throw new XcfTypeInitializationException(
                            "No server with name '" + remoteServerName
                                    + "' found.", e);
                } else {
                    logger.warn("Right now no remote server for name '"
                            + remoteServerName + "' could be created. "
                            + "Initialization will be delayed.", e);
                    remoteServersByName.put(remoteServerName, null);
                }

            }

        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void initializeAndManage(RemoteServerReconnectable reconnectable)
            throws XcfTypeInitializationException {

        logger.info("Got request to initialize and manage new reconnectable: "
                + reconnectable + " for remote servers '"
                + reconnectable.getRemoteServerNames() + "'.");

        synchronized (remoteServerLock) {

            for (Entry<String, ErrorHandlingPolicy> serverEntry : reconnectable
                    .getRemoteServerNames().entrySet()) {

                String remoteServerName = serverEntry.getKey();
                ErrorHandlingPolicy errorPolicy = serverEntry.getValue();

                logger.debug("Initializing remote server " + remoteServerName
                        + " for reconnectable " + reconnectable
                        + " with error policy " + errorPolicy);

                // create new remote server if needed
                if (!remoteServersByName.containsKey(remoteServerName)) {

                    createRemoteServer(remoteServerName);

                    logger.warn("There was no remote server for name '"
                            + remoteServerName
                            + "' initialized. Creation was needed.");

                }

                assert remoteServersByName.containsKey(remoteServerName);

                // decide which error handling policy is requested for this
                // server by this client
                switch (errorPolicy) {
                case SELF_HEALING:
                    initializeDecoratorClient(reconnectable, remoteServerName);
                    break;
                case NOTIFYING:
                    initializeDirectClient(reconnectable, remoteServerName);
                    break;
                default:
                    logger.error("Unknown error handling policy " + errorPolicy
                            + " for server " + remoteServerName
                            + ". This server will not be initialized.");
                    break;
                }

            }

        }

    }

    /**
     * Initializes a reconnectable that wants to be informed directly about
     * connections errors. Assumes that {@link #remoteServerLock} is held.
     * 
     * @param reconnectable
     *            reconnectable to initialize
     * @param remoteServerName
     *            name of the requested remote server
     */
    private void initializeDirectClient(
            RemoteServerReconnectable reconnectable, String remoteServerName) {

        // initialize reconnectable if possible
        if (remoteServersByName.get(remoteServerName) != null) {
            reconnectable
                    .resetRemoteServer(new ConnectionErrorNotifyingRemoteServer(
                            remoteServersByName.get(remoteServerName), this));
            logger.info("Initialized reconnectable " + reconnectable
                    + " with direct remote server.");
        } else {
            logger.warn("There was no accessible remote server "
                    + "right now. Connection will be postponed "
                    + "until a remote server with the requested "
                    + "name can be accessed.");
        }

        // add reconnectable to list of directly notified reconnectables
        if (!directNotifiers.containsKey(remoteServerName)) {
            directNotifiers.put(remoteServerName,
                    new HashSet<RemoteServerReconnectable>());
        }
        directNotifiers.get(remoteServerName).add(reconnectable);

    }

    /**
     * Initializes a reconnectable that wants to use the decorator. Assumes that
     * {@link #remoteServerLock} is held.
     * 
     * @param reconnectable
     *            reconnectable to initialize
     * @param remoteServerName
     *            name of the requested remote server
     */
    private void initializeDecoratorClient(
            RemoteServerReconnectable reconnectable, String remoteServerName) {

        // setup a new decorator for this reconnectable
        ConnectionErrorResistantRemoteServer decorator =
                new ConnectionErrorResistantRemoteServer(this);
        if (remoteServersByName.get(remoteServerName) != null) {
            decorator.resetRemoteServer(remoteServersByName
                    .get(remoteServerName));
        } else {
            logger.warn("There was no accessible remote server "
                    + "right now. Connection will be postponed "
                    + "until a remote server with the requested "
                    + "name can be accessed.");
        }

        // and add this decorator to the list of managed decorators
        if (!decoratorsByServerName.containsKey(remoteServerName)) {
            decoratorsByServerName.put(remoteServerName,
                    new HashSet<ConnectionErrorResistantRemoteServer>());
        }
        decoratorsByServerName.get(remoteServerName).add(decorator);

        // and give the decorator to the reconnectable
        logger.debug("setting remote server " + remoteServerName
                + " for reconnectable " + reconnectable);
        reconnectable.resetRemoteServer(decorator);

    }

    /**
     * Marks a remote server as being disconnected.
     * 
     * @param remoteServerName
     */
    void markRemoteServerAsDisconnected(String remoteServerName) {

        synchronized (remoteServerLock) {

            assert remoteServersByName.containsKey(remoteServerName);

            // avoid race conditions
            if (remoteServersByName.get(remoteServerName) != null) {
                remoteServersByName.put(remoteServerName, null);
            }

            logger.info("Marked remote server '" + remoteServerName
                    + "' as disconnected.");

        }

    }

}
