package de.unibi.citec.clf.bonsai.xcf.actuators;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.unibi.citec.clf.bonsai.actuators.PoseActuator;
import de.unibi.citec.clf.bonsai.core.CommunicationException;
import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.SyntaxException;
import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;

/**
 * Small helper class to invoke standard postures on the robot.
 * 
 * @author Ingo Luetkebohle <iluetkeb@techfak.uni-bielefeld.de>
 */
public class XcfPoseActuator implements RemoteServerReconnectable, PoseActuator {

	/** Server for communication with BWorld. */
	private String remoteServerName = "";
	private Object remoteServerLock = new Object();
	private RemoteServer server;

	/**
	 * Constructor.
	 * 
	 * @param remoteServerName
	 *            name of the remote server used to control the pose
	 */
	public XcfPoseActuator(String remoteServerName) {
		this.remoteServerName = remoteServerName;

		// create new template instance
		Element rootElement = new Element("motionTask");
		Element element = new Element("pose");
		template = new Document(rootElement);
		template.getRootElement().appendChild(element);
	}

	/** Call the server and return the result. */
	protected String call(final Document xml) throws UserException {
		try {
			XOPData out = new XOPData(xml);
			XOPData result;
			synchronized (remoteServerLock) {
				result = server.callMethod("setParametrizedPose", out);
			}
			return result.getDocumentAsText();
		} catch (MethodNotFoundException ex) {
			// this should not happen, its a runtime exception
			Logger.getLogger(XcfPoseActuator.class.getName()).log(Level.SEVERE,
					null, ex);
			throw new RuntimeException("Server is missing required method", ex);
		}
	}

	// set up the standard template for sending data
	private final Document template;

	protected final void setName(Element e, String name) {
		e.addAttribute(new Attribute("name", name));
	}

	protected final void setDouble(Element e, String attrName, double d) {
		e.addAttribute(new Attribute(attrName, Double.toString(d)));
	}

	/**
	 * Fills in the given arguments into the template and returns it. Makes a
	 * copy of the template before doing so.
	 */
	protected Document createTargetXML(String name, double x, double y, double z) {
		Document d = (Document) template.copy();
		Element tag = d.getRootElement().getFirstChildElement("pose");
		setName(tag, name);
		setDouble(tag, "x", x);
		setDouble(tag, "y", y);
		setDouble(tag, "z", z);
		return d;
	}

	protected Document createAngleXML(String name, double angle) {
		Document d = (Document) template.copy();
		Element tag = d.getRootElement().getFirstChildElement("pose");
		setName(tag, name);
		setDouble(tag, "angle", angle);
		return d;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void pointAt(double x, double y, double z)
			throws CommunicationException {
		try {
			call(createTargetXML("pointTo", x, y, z));
		} catch (UserException e) {
			throw new CommunicationException(e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void lookAt(double x, double y, double z)
			throws CommunicationException {
		try {
			call(createTargetXML("lookAt", x, y, z));
		} catch (UserException e) {
			throw new CommunicationException(e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void rotateBodyTo(double degrees) throws CommunicationException {
		try {
			call(createAngleXML("rotateBody", degrees));
		} catch (UserException e) {
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void resetToBase() {
		try {
			XOPData d = new XOPData();
			d.setDocument("<?xml version=\"1.0\"?><foo/>");
			synchronized (remoteServerLock) {
				server.callMethod("resetToBase", d);
			}
		} catch (SyntaxException ex) {
			Logger.getLogger(XcfPoseActuator.class.getName()).log(Level.SEVERE,
					null, ex);
		} catch (MethodNotFoundException ex) {
			Logger.getLogger(XcfPoseActuator.class.getName()).log(Level.SEVERE,
					null, ex);
		} catch (UserException ex) {
			Logger.getLogger(XcfPoseActuator.class.getName()).log(Level.SEVERE,
					null, ex);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
		Map<String, ErrorHandlingPolicy> servers = new HashMap<String, ErrorHandlingPolicy>();
		servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
		return servers;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void resetRemoteServer(RemoteServer remoteServer) {
		assert remoteServer.getServerName().equals(remoteServerName);
		synchronized (remoteServerLock) {
			this.server = remoteServer;
		}
	}

	/** Get pose actuator manager for the robot with the given name. */
	// curry drive-by
	// public static ActuatorManager<PoseActuator> getActuatorManager(String
	// name) {
	// return new ActuatorManager<PoseActuator>(new
	// ManagerStatePoseActuatorUnavailable(name));
	// }
}
