package de.unibi.citec.clf.bonsai.xcf.actuators;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Timer;
import java.util.Vector;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.SyntaxException;
import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Elements;
import nu.xom.ParsingException;

import org.apache.log4j.Logger;

import de.unibi.citec.clf.bonsai.actuators.PoseActuatorTobi;
import de.unibi.citec.clf.bonsai.core.Actuator;
import de.unibi.citec.clf.bonsai.util.arm.ArmStatus;
import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.grasp.GraspReturnType;
import de.unibi.airobots.btl.data.grasp.KatanaGripperData;
import de.unibi.airobots.btl.data.grasp.KatanaPoseData;
import de.unibi.airobots.btl.data.grasp.PoseData;
import de.unibi.airobots.btl.data.object.ObjectShapeData;
import de.unibi.airobots.btl.data.object.ObjectShapeList;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;

/**
 * Pose Actuator for Tobi Movements have to define in the vdemo-Config File.
 * 
 * @author ttoenige
 */
public class XcfPoseActuatorTobi implements PoseActuatorTobi,
        RemoteServerReconnectable, Actuator {

    private Logger logger = Logger.getLogger(getClass());

    private String remoteServerName = "";
    private Object remoteServerLock = new Object();
    private RemoteServer server;

    private final Document postureTemplate;
    private final Document gotoTemplate;

    private Timer gripperScheduler;

    // private volatile boolean isClosing = false;

    /**
     * Constructor.
     * 
     * @param remoteServerName
     *            name of the remote server to connect to
     */
    public XcfPoseActuatorTobi(String remoteServerName) {

        this.remoteServerName = remoteServerName;

        Element postureRootElement = new Element("POSTURE");
        postureTemplate = new Document(postureRootElement);

        Element gotoRootElement = new Element("GOTO");
        gotoTemplate = new Document(gotoRootElement);

        gripperScheduler = new Timer();
    }

    /** Call the server and return the result. */
    private XOPData call(final Document xml, String method) throws IOException {
        try {
            XOPData out = new XOPData(xml);
            synchronized (remoteServerLock) {
                return server.callMethod(method, out);
            }
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (MethodNotFoundException ex) {
            logger.error(ex);
            throw new IOException(ex);

        }
    }

    private void setDouble(Element e, String attrName, double d) {
        e.addAttribute(new Attribute(attrName, Double.toString(d)));
    }

    private void setInt(Element e, String attrName, int i) {
        e.addAttribute(new Attribute(attrName, Integer.toString(i)));
    }

    private Document createMoveJointXML(int joint, double value) {
        Document d = (Document) postureTemplate.copy();
        Element element = new Element("POSITION");
        setInt(element, "joint", joint);
        setDouble(element, "value", value);
        d.getRootElement().appendChild(element);
        return d;
    }

    private Document createGotoXML(double px, double py, double pz, double nx,
            double ny, double nz, double ax, double ay, double az, double ox,
            double oy, double oz) {
        Document d = (Document) gotoTemplate.copy();
        Element pElement = new Element("P");
        setDouble(pElement, "x", px);
        setDouble(pElement, "y", py);
        setDouble(pElement, "z", pz);
        d.getRootElement().appendChild(pElement);
        Element nElement = new Element("N");
        setDouble(nElement, "x", nx);
        setDouble(nElement, "y", ny);
        setDouble(nElement, "z", nz);
        d.getRootElement().appendChild(nElement);
        Element aElement = new Element("A");
        setDouble(aElement, "x", ax);
        setDouble(aElement, "y", ay);
        setDouble(aElement, "z", az);
        d.getRootElement().appendChild(aElement);
        Element oElement = new Element("O");
        setDouble(oElement, "x", ox);
        setDouble(oElement, "y", oy);
        setDouble(oElement, "z", oz);
        d.getRootElement().appendChild(oElement);
        return d;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void moveJoint(int joint, double value) throws IOException {
        logger.debug("moveJoint");
        call(createMoveJointXML(joint, value), "moveJoints");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public LinkedList<Double> listJoints() throws IOException {
        logger.debug("listAngles");
        XOPData d = new XOPData();
        LinkedList<Double> joints = new LinkedList<Double>();
        try {
            d.setDocument("<?xml version=\"1.0\"?><foo/>");
            Document result = call(d.getDocument(), "listAngles").getDocument();

            Elements elements = result.getRootElement().getChildElements();
            int size = elements.size();

            for (int i = 0; i < size; i++) {
                joints.add(Double.parseDouble(elements.get(i)
                        .getAttributeValue("value")));
            }
            return joints;

        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean goTo(double px, double py, double pz, double nx, double ny,
            double nz, double ax, double ay, double az, double ox, double oy,
            double oz) {
        try {
            logger.debug("goTo");
            call(createGotoXML(px, py, pz, nx, ny, nz, ax, ay, az, ox, oy, oz),
                    "goTo").getDocument();
            return true;
        } catch (IOException e) {
            logger.error("Problem in goTo function!");
            logger.error(e);
            return false;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Vector<String> listMovements() throws IOException {
        logger.debug("listMovements");
        XOPData d = new XOPData();
        Vector<String> movements = new Vector<String>();
        try {
            d.setDocument("<?xml version=\"1.0\"?><foo/>");
            Document result =
                    call(d.getDocument(), "listMovements").getDocument();

            Elements poses =
                    result.getRootElement().getFirstChildElement("POSES")
                            .getChildElements();
            for (int i = 0; i < poses.size(); i++) {
                movements.add(poses.get(i).getAttributeValue("name"));
            }

            Elements actions =
                    result.getRootElement().getFirstChildElement("ACTIONS")
                            .getChildElements();
            for (int j = 0; j < actions.size(); j++) {
                movements.add(actions.get(j).getAttributeValue("name"));
            }

        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
        return movements;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setMovements(String name) throws IOException {
        logger.debug("setMovements");
        XOPData d = new XOPData();
        try {
            d.setDocument("<MOVEMENT name=\"" + name + "\"/>");
            call(d.getDocument(), "setMovement").getDocument();
        } catch (SyntaxException e) {
            //logger.error(e); TODO: thrown nearly every call. So disable loggin for better log style.
            throw new IOException(e);
        }
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void setCarryMovements(String name) throws IOException {
        logger.debug("setMovements");
        XOPData d = new XOPData();
        try {
            d.setDocument("<MOVEMENT name=\"" + name + "\"/>");
            call(d.getDocument(), "setCarryMovement").getDocument();
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void executeGraspMovement () throws IOException {
        logger.debug("exectueGraspMovement");
        XOPData d = new XOPData();
        try {
            d.setDocument("<EXECUTEGRASPMOVEMENT />");
            call(d.getDocument(), "executeGraspMovement").getDocument();
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void openGripper() throws IOException {
        logger.debug("openGripper");
        try {
            gripperScheduler.cancel();
        } catch (IllegalStateException e) {
            logger.warn("Illegal state", e);
        }
        // isClosing = false;
        XOPData d = new XOPData();
        try {
            d.setDocument("<?xml version=\"1.0\"?><foo/>");
            call(d.getDocument(), "openGripper");
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void closeGripper() throws IOException {
        logger.debug("closeGripper");
        XOPData d = new XOPData();
        try {
            d.setDocument("<foo/>");
            call(d.getDocument(), "closeGripper");
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }

        /*
         * if (!isClosing) { try { gripperScheduler.cancel(); } catch
         * (IllegalStateException e) { logger.warn("Illegal state", e); }
         * gripperScheduler = new Timer();
         * gripperScheduler.scheduleAtFixedRate(new TimerTask() {
         * 
         * public void run() { try { closeGripper(); } catch (IOException e) {
         * logger.error("Error while manipulating gripper"); } } }, CLOSING_WAIT
         * // Startverzögerung , CLOSING_WAIT // Dauer ); isClosing = true; }
         */
    }

    /**
     * 
     * {@inheritDoc}
     */
    public void closeGripperByForce() throws IOException {
        logger.debug("closeGripperbyForce");
        XOPData d = new XOPData();
        try {
            d.setDocument("<foo/>");
            call(d.getDocument(), "closeGripperByForce");
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    public void openGripperWhenTouching(int waitSeconds) throws IOException {
        logger.debug("open Gripper when touching");
        XOPData d = new XOPData();
        
        try {
            Thread.sleep(waitSeconds*1000);
            d.setDocument("<foo/>");
            call(d.getDocument(), "openGripperWhenTouching");
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (InterruptedException e) {
            logger.fatal(e.getMessage());
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public KatanaGripperData getGipperSensorData() throws IOException {
        KatanaGripperData parsed = new KatanaGripperData();
        logger.debug("get sensor data");
        XOPData d = new XOPData();
        try {
            d.setDocument("<foo/>");
            Document callResult;

            callResult =
                    call(d.getDocument(), "getGripperSensors").getDocument();

            KatanaGripperData.fromElement(callResult.getRootElement(), parsed);

        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
        return parsed;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean goTo(KatanaPoseData position, boolean linear)
            throws IOException {
        Element goToElement = new Element("GOTO");
        goToElement.addAttribute(new Attribute("name", "euler"));
        if (linear) {
            goToElement.addAttribute(new Attribute("linear", "1"));
        } else {
            goToElement.addAttribute(new Attribute("linear", "0"));
        }

        Document positionDoc =
                new Document(new Element(KatanaPoseData.getBaseTag()));
        position.fillInto(positionDoc.getRootElement());

        Element posElem = new Element(positionDoc.getRootElement());
        goToElement.appendChild(posElem);

        logger.info(goToElement.toXML());

        //goToElement.appendChild(positionDoc.getRootElement());

        logger.info(goToElement.toXML());

        XOPData d = new XOPData();

        try {

            d.setDocument(goToElement.toXML());

            Document callResult;
            callResult = call(d.getDocument(), "goTo").getDocument();

            String success = callResult.getRootElement().getValue();
            if (success.equals("true")) {
                return true;
            } else {
                return false;
            }

        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }

    }

    public ArmStatus executeStrategy(PoseData target) throws IOException {
        Document doc = new Document(new Element(PoseData.getBaseTag()));
        target.fillInto(doc.getRootElement());

        logger.info(doc.toXML());

        XOPData d = new XOPData();
        try {
            d.setDocument(doc.toXML());
            Document callResult;
            callResult =
                    call(d.getDocument(), "executeGraspStrategy").getDocument();

            ArmStatus armStatus =
                    new ArmStatus(
                            Boolean.parseBoolean(callResult.getRootElement()
                                    .getAttribute("grasp").getValue()), Boolean
                                    .parseBoolean(callResult.getRootElement()
                                            .getAttribute("approach")
                                            .getValue()), Boolean
                                    .parseBoolean(callResult.getRootElement()
                                            .getAttribute("lift").getValue()),
                            Boolean.parseBoolean(callResult.getRootElement()
                                    .getAttribute("carry").getValue()));

            logger.debug("Approach: " + armStatus.isApproachPositionReached()
                    + " Lift: " + armStatus.isLiftPositionReached()
                    + " Grasp: " + armStatus.isGraspingObject() + " Carry: "
                    + armStatus.isCarryPositionReached());

            return armStatus;

        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        assert remoteServer.getServerName().equals(remoteServerName);
        synchronized (remoteServerLock) {
            this.server = remoteServer;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public KatanaPoseData getPosition() throws IOException {
        logger.debug("getPosition");
        XOPData d = new XOPData();
        KatanaPoseData kpd = new KatanaPoseData();
        try {
            d.setDocument("<?xml version=\"1.0\"?><foo/>");
            Document result =
                    call(d.getDocument(), "getPosition").getDocument();

            Element poses =
                result.getRootElement().getFirstChildElement("P");
            kpd.setX(Double.parseDouble(poses.getAttributeValue("x")),LengthUnit.MILLIMETER);
            kpd.setY(Double.parseDouble(poses.getAttributeValue("y")),LengthUnit.MILLIMETER);
            kpd.setZ(Double.parseDouble(poses.getAttributeValue("z")),LengthUnit.MILLIMETER);
            poses =
                result.getRootElement().getFirstChildElement("ANGLE");
            kpd.setPhi(Double.parseDouble(poses.getAttributeValue("phi")),AngleUnit.RADIAN);
            kpd.setPsi(Double.parseDouble(poses.getAttributeValue("psi")),AngleUnit.RADIAN);
            kpd.setTheta(Double.parseDouble(poses.getAttributeValue("theta")),AngleUnit.RADIAN);

        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        }
        return kpd;
    }

    /**
     * {@inheritDoc}
     */
	@Override
	public void motorsOff() throws IOException {
		logger.debug("motorsOff");
		try {
			server.callMethod("motorsOff", null);
		} catch (MethodNotFoundException e) {
			logger.fatal(e.getMessage());
		} catch (UserException e) {
			logger.fatal(e.getMessage());
		}
	}

	/**
     * {@inheritDoc}
     */
	@Override
	public void motorsOn() throws IOException {
		logger.debug("motorsOn");
		try {
			server.callMethod("motorsOn", null);
		} catch (MethodNotFoundException e) {
			logger.fatal(e.getMessage());
		} catch (UserException e) {
			logger.fatal(e.getMessage());
		}
		
	}

		
	@Override
	public GraspReturnType graspObjectAt(ObjectShapeData objectToGrasp, ObjectShapeList otherObjects)
			throws IOException {
		logger.debug("graspObjectAt called");
        Element goToElement = new Element("GRASPOBJECTAT");

        Document positionDoc =
                new Document(new Element(ObjectShapeData.getBaseTag()));
        objectToGrasp.fillInto(positionDoc.getRootElement());

        Element posElem = new Element(positionDoc.getRootElement());
        goToElement.appendChild(posElem);
        
        Document positionDoc2 =
                new Document(new Element(ObjectShapeList.getBaseTag()));
        otherObjects.fillInto(positionDoc2.getRootElement());

        Element posElem2 = new Element(positionDoc2.getRootElement());
        goToElement.appendChild(posElem2);

        logger.info(goToElement.toXML());
		
	
        XOPData d = new XOPData();
        GraspReturnType grt = new GraspReturnType();

        try {

            d.setDocument(goToElement.toXML());

            Document callResult;
            callResult = call(d.getDocument(), "graspObjectAt").getDocument();       
            GraspReturnType.fromElement(callResult.getRootElement(), grt);

        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (ParsingException e) {
			logger.error(e);
		}
        
        return grt;
        
	}
	
	@Override
	public GraspReturnType isObjectGraspable(ObjectShapeData objectToGrasp, ObjectShapeList otherObjects)
			throws IOException {
		logger.debug("isObjectGraspable called");
        Element goToElement = new Element("ISOBJECTGRASPABLE");

        Document positionDoc =
                new Document(new Element(ObjectShapeData.getBaseTag()));
        objectToGrasp.fillInto(positionDoc.getRootElement());

        Element posElem = new Element(positionDoc.getRootElement());
        goToElement.appendChild(posElem);
        
        Document positionDoc2 =
                new Document(new Element(ObjectShapeList.getBaseTag()));
        otherObjects.fillInto(positionDoc2.getRootElement());

        Element posElem2 = new Element(positionDoc2.getRootElement());
        goToElement.appendChild(posElem2);

        XOPData d = new XOPData();
        GraspReturnType grt = new GraspReturnType();

        try {

            d.setDocument(goToElement.toXML());

            Document callResult;                   
            callResult = call(d.getDocument(), "isObjectGraspable").getDocument();     
            GraspReturnType.fromElement(callResult.getRootElement(), grt);
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (ParsingException e) {
			logger.error(e);
		}
        
        return grt;
        
	}
	
	@Override
	public GraspReturnType wipingMovement(KatanaPoseData height)
			throws IOException {
		logger.debug("wipingMovement called");
        Element goToElement = new Element("WIPINGMOVEMENT");

        Document positionDoc =
                new Document(new Element(ObjectShapeData.getBaseTag()));
        height.fillInto(positionDoc.getRootElement());

        Element posElem = new Element(positionDoc.getRootElement());
        goToElement.appendChild(posElem);
        
        XOPData d = new XOPData();
        GraspReturnType grt = new GraspReturnType();

        try {

            d.setDocument(goToElement.toXML());

            Document callResult;                   
            callResult = call(d.getDocument(), "wipingMovement").getDocument();     
            GraspReturnType.fromElement(callResult.getRootElement(), grt);
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (ParsingException e) {
			logger.error(e);
		}
        
        return grt;
        
	}
	
	@Override
	public GraspReturnType placeObjectAt(KatanaPoseData position)
			throws IOException {
		logger.debug("isObjectGraspable called");
        Element goToElement = new Element("PLACEOBJECTAT");

        Document positionDoc =
                new Document(new Element(KatanaPoseData.getBaseTag()));
        position.fillInto(positionDoc.getRootElement());

        Element posElem = new Element(positionDoc.getRootElement());
        goToElement.appendChild(posElem);
        
                XOPData d = new XOPData();
        GraspReturnType grt = new GraspReturnType();

        try {

            d.setDocument(goToElement.toXML());

            Document callResult;                   
            callResult = call(d.getDocument(), "placeObjectAt").getDocument();     
            GraspReturnType.fromElement(callResult.getRootElement(), grt);
        } catch (SyntaxException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (ParsingException e) {
			logger.error(e);
		}
        
        return grt;
        
	}

	@Override
	public void freeze() throws IOException {
		logger.debug("freeze");
		try {
			server.callMethod("freeze", null);
		} catch (MethodNotFoundException e) {
			logger.error(e);
            throw new IOException(e);
		} catch (UserException e) {
			logger.error(e);
            throw new IOException(e);
		}
	}
	
	@Override
	public void unblock() throws IOException {
		logger.debug("unblock");
		try {
			server.callMethod("unblock", null);
		} catch (MethodNotFoundException e) {
			logger.error(e);
            throw new IOException(e);
		} catch (UserException e) {
			logger.error(e);
            throw new IOException(e);
		}
	}

		@Override
		public GraspReturnType openDoor(KatanaPoseData position)
				throws IOException {
			logger.debug("findLeverAndMoveDown called");
	        Element goToElement = new Element("FINDLEVERANDMOVEDOWN");

	        Document positionDoc =
	                new Document(new Element(KatanaPoseData.getBaseTag()));
	        position.fillInto(positionDoc.getRootElement());

	        Element posElem = new Element(positionDoc.getRootElement());
	        goToElement.appendChild(posElem);
	        
	                XOPData d = new XOPData();
	        GraspReturnType grt = new GraspReturnType();

	        try {

	            d.setDocument(goToElement.toXML());

	            Document callResult;                   
	            callResult = call(d.getDocument(), "findLeverAndMoveDown").getDocument();     
	            GraspReturnType.fromElement(callResult.getRootElement(), grt);
	        } catch (SyntaxException e) {
	            logger.error(e);
	            throw new IOException(e);
	        } catch (ParsingException e) {
				logger.error(e);
			}
	        
	        return grt;
	        
		}

		@Override
		public GraspReturnType joggingMovement(KatanaPoseData pose)
				throws IOException {
			logger.debug("joggingMovement called");
	        Element goToElement = new Element("JOGGINGMOVEMENT");

	        Document positionDoc =
	                new Document(new Element(KatanaPoseData.getBaseTag()));
	        pose.fillInto(positionDoc.getRootElement());

	        Element posElem = new Element(positionDoc.getRootElement());
	        goToElement.appendChild(posElem);
	        
	                XOPData d = new XOPData();
	        GraspReturnType grt = new GraspReturnType();

	        try {

	            d.setDocument(goToElement.toXML());

	            Document callResult;                   
	            callResult = call(d.getDocument(), "joggingMovement").getDocument();     
	            GraspReturnType.fromElement(callResult.getRootElement(), grt);
	        } catch (SyntaxException e) {
	            logger.error(e);
	            throw new IOException(e);
	        } catch (ParsingException e) {
				logger.error(e);
			}
	        
	        return grt;
		}
}