package de.unibi.citec.clf.bonsai.xcf.actuators;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import nu.xom.Document;
import nu.xom.Element;

import org.apache.log4j.Logger;

import de.unibi.citec.clf.bonsai.actuators.IsrActuator;
import de.unibi.citec.clf.bonsai.actuators.SpeechActuator;
import de.unibi.citec.clf.bonsai.actuators.SpeechActuatorListener;
import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable;

/**
 * Actuator which can be used to access the speech synthesizer.
 * 
 * @author lschilli
 * @author jwienke
 * @author arothert
 */
public class XcfSpeechActuator extends SpeechActuator implements
		RemoteServerReconnectable {

	private Logger logger = Logger.getLogger(getClass());
	private String remoteServerName = "";
	private final Object remoteServerLock = new Object();
	private RemoteServer remoteServer = null;
	private static final int ISR_START_DB_HIGH = 60;
	private static final int ISR_UTT_DB_HIGH = 60;
	private static final int ISR_UTT_DB_LOW = 35;
	private static final int ISR_START_DB_LOW = 30;
	private IsrActuator isrActuator;

	/**
	 * Constructor.
	 * 
	 * @param remoteServerName
	 *            name of the remote server used to produce speech
	 */
	public XcfSpeechActuator(String remoteServerName) {
		this.remoteServerName = remoteServerName;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void say(String text, boolean async) throws IOException {
		logger.info("Biron: "+text);
		try {
			isrActuator.setThresholds(ISR_START_DB_HIGH, ISR_UTT_DB_HIGH);
		} catch (Throwable e) {
			logger.warn("No IsrActuator found");
			logger.warn("No Treshold adapting");
		}
		if (isAllowedSayChar(text)) {

			synchronized (remoteServerLock) {
				if (remoteServer == null) {
					throw new IllegalStateException("No remote server set.");
				}
			}

			Element elem = new Element("raw");
			elem.appendChild(text);
			Document doc = new Document(elem);
			try {
				callListeners(text);
				if (async) {
					synchronized (remoteServerLock) {
						remoteServer.callMethod("async_say_raw", new XOPData(
								doc));
					}
				} else {
					synchronized (remoteServerLock) {
						remoteServer.callMethod("say_raw", new XOPData(doc));
					}
				}
			} catch (MethodNotFoundException e) {
				logger.error(e);
				throw new IOException(e);
			} catch (UserException e) {
				logger.error(e);
				throw new IOException(e);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void say(String text) throws IOException {
		say(text, false);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
		Map<String, ErrorHandlingPolicy> servers = new HashMap<String, ErrorHandlingPolicy>();
		servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
		return servers;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void resetRemoteServer(RemoteServer remoteServer) {
		assert remoteServer.getServerName().equals(remoteServerName);
		synchronized (remoteServerLock) {
			this.remoteServer = remoteServer;
		}
	}

	@Override
	public void sayAccentuated(String accentuatedText, boolean async)
			throws IOException {
		try {
			isrActuator.setThresholds(ISR_START_DB_HIGH, ISR_UTT_DB_HIGH);
		} catch (Throwable e) {
			logger.warn("No IsrActuator found");
			logger.warn("No Treshold adapting");
		}
		if (isAllowedSayChar(accentuatedText)) {

			synchronized (remoteServerLock) {
				if (remoteServer == null) {
					throw new IllegalStateException("No remote server set.");
				}
			}

			Element elem = new Element("accentuatedText");
			Element text = new Element("text");
			Element bool = new Element("bool");

			text.appendChild(accentuatedText);
			bool.appendChild(Boolean.toString(async));
			elem.appendChild(text);
			elem.appendChild(bool);

			Document doc = new Document(elem);
			
			try {
				callListeners(accentuatedText);
				synchronized (remoteServerLock) {
					remoteServer.callMethod("say_accented", new XOPData(doc));
				}
			} catch (MethodNotFoundException e) {
				logger.error(e);
				throw new IOException(e);
			} catch (UserException e) {
				logger.error(e);
				throw new IOException(e);
			}
		}
	}

	@Override
	public void sayAccentuated(String accentuatedText, boolean async,
			String prosodyConfig) throws IOException {
		try {
			isrActuator.setThresholds(ISR_START_DB_HIGH, ISR_UTT_DB_HIGH);
		} catch (Throwable e) {
			logger.warn("No IsrActuator found");
			logger.warn("No Treshold adapting");
		}
		if (isAllowedSayChar(accentuatedText)) {

			synchronized (remoteServerLock) {
				if (remoteServer == null) {
					throw new IllegalStateException("No remote server set.");
				}
			}

			Element elem = new Element("accentuatedText");
			Element text = new Element("text");
			Element bool = new Element("bool");
			Element prosody = new Element("prosody");

			text.appendChild(accentuatedText);
			bool.appendChild(Boolean.toString(async));
			prosody.appendChild(prosodyConfig);

			elem.appendChild(text);
			elem.appendChild(bool);
			elem.appendChild(prosody);

			Document doc = new Document(elem);
			try {
				callListeners(accentuatedText);
				synchronized (remoteServerLock) {
					remoteServer.callMethod("say_accented", new XOPData(doc));
				}
			} catch (MethodNotFoundException e) {
				logger.error(e);
				throw new IOException(e);
			} catch (UserException e) {
				logger.error(e);
				throw new IOException(e);
			}
		}
	}
	
	@Override
	public void setIsrActuator(IsrActuator isrAct) {
		isrActuator = isrAct;
	}

	private static void callListeners(String text) {
		for (SpeechActuatorListener listener : listeners) {
			listener.newUtterance(text);
		}
	}

	@Override
	public void sayAccentuated(String accented_text) throws IOException {
		sayAccentuated(accented_text, false);
	}

	@Override
	public void sayAccentuated(String accented_text, String prosodyConfig)
			throws IOException {
		sayAccentuated(accented_text, false, prosodyConfig);
	}

	/**
	 * check if string entails signs which through a null-pointer exception
	 * 
	 * @param str
	 *            say string which should be spoken
	 * @return true if the string is allowed else false
	 */
	private boolean isAllowedSayChar(String str) {

		Pattern p = Pattern.compile("[ ]*?[.,!?;:()]+");
		Matcher m = p.matcher(str);

		if (m.matches() || str.equals("")) {
			logger.error("String to say is one of the following symbols: ,.!?():; see: "
					+ str);
			return false;
		}
		return true;
	}
}
