package de.unibi.citec.clf.bonsai.xcf.sensors;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.Sensor;
import de.unibi.citec.clf.bonsai.core.SensorListener;
import de.unibi.citec.clf.bonsai.xcf.BoundSynchronizedQueue;
import de.unibi.citec.clf.bonsai.xcf.SubscriberReconnectable;
import de.unibi.airobots.btl.Type;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Set;
import net.sf.xcf.Subscriber;
import net.sf.xcf.event.PublishEvent;
import net.sf.xcf.event.PublishEventAdapter;
import net.sf.xcf.event.XcfEvent;
import net.sf.xcf.transport.XOPData;
import nu.xom.Document;
import org.apache.log4j.Logger;

/**
 * A simple {@link Sensor} for BTL based types that uses their static
 * <code>fromDocument</code> method to extract the data. This allows this sensor
 * to be used for every XCF-publisher that provides a BTL parseable result.
 * 
 * @author jwienke
 * @param <T>
 *            btl type returned by this sensor
 */
public class XcfBtlPublisherSensor<T extends Type> extends PublishEventAdapter
        implements Sensor<T>, SubscriberReconnectable {

    private String subscriberName = "";

    private Logger logger = Logger.getLogger(getClass());

    private Class<T> typeClass;
    private Subscriber subscriber = null;
    private BoundSynchronizedQueue<T> queue = new BoundSynchronizedQueue<T>() {

        @Override
        public T convertFromXcfEvent(XcfEvent event) {
            // we don't use this method
            return null;
        }

    };
    private Method fromDocumentMethod = null;
    private Set<SensorListener<T>> listeners = new HashSet<SensorListener<T>>();

    /**
     * Constructs a new instance of this sensor for the given BTL type based on
     * a publisher.
     * 
     * @param typeClass
     *            type to return by this sensor.
     * @param subscriberName
     *            name of the subscriber this sensor will work on
     * @throws IllegalArgumentException
     *             the given type does not contain a static fromDocument method.
     */
    public XcfBtlPublisherSensor(Class<T> typeClass, String subscriberName)
            throws IllegalArgumentException {

        this.typeClass = typeClass;
        this.subscriberName = subscriberName;

        // find fromDocument method for the btl type
        try {
            fromDocumentMethod =
                    typeClass.getMethod("fromDocument", new Class[] {
                            Document.class, Class.class });
        } catch (SecurityException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'fromDocument' method in the given type "
                            + typeClass + ".", e);
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'fromDocument' method in the given type "
                            + typeClass + ".", e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetSubscriber(Subscriber subscriber) {

        logger.info("Reconnecting to new subscriber: " + subscriber);

        if (this.subscriber != null) {
            logger.debug("There was an old subscriber for this sensor. "
                    + "First de-registering from this subscriber.");
            this.subscriber.removeListener(this);
        }

        this.subscriber = subscriber;
        this.subscriber.addListener(this);

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void addSensorListener(SensorListener<T> listener) {
        listeners.add(listener);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Class<T> getDataType() {
        return typeClass;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public T readLast() throws IOException, InterruptedException {
        logger.debug("Queue size: " + queue.getSize());
        return queue.next();
    }

    /**
     * Safely extracts the BTL type from the document found in the
     * {@link XOPData}.
     * 
     * @param data
     *            data containing the document
     * @return extracted BTL type
     * @throws IOException
     *             error extracting BTL type instance
     */
    @SuppressWarnings("unchecked")
    private T safeExtractType(XOPData data) throws IOException {

        Document doc = data.getDocument();
        if (doc == null) {
            throw new IOException("Document empty in XOPData: " + data);
        }

        try {
            Object rawResult = fromDocumentMethod.invoke(null, doc, typeClass);
            if (rawResult == null) {
                throw new IOException(
                        "fromDocument method has wrong signature "
                                + "or simply returned null.");
            }
            postProcessType(data, (T) rawResult);
            return (T) rawResult;
        } catch (IllegalArgumentException e) {
            throw new IOException(e);
        } catch (IllegalAccessException e) {
            throw new IOException(e);
        } catch (InvocationTargetException e) {
            throw new IOException("Error parsing the document.", e);
        } catch (ExceptionInInitializerError e) {
            throw new IOException(e);
        }

    }

    /**
     * Hook method that may be implemented in subclasses to post process the
     * simple BTL based de-serialization of the data type.
     * 
     * @param data
     *            {@link XOPData} for the event
     * @param extractedType
     *            type already extract with BTL <code>fromDocument</code>
     * @throws IOException
     *             the post processing failed and returning the type to callers
     *             would be illegal
     */
    protected void postProcessType(XOPData data, T extractedType)
            throws IOException {

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public T readLast(long timeout) throws IOException, InterruptedException {
        return queue.nextCached(timeout);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void removeSensorListener(SensorListener<T> listener) {
        listeners.remove(listener);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void handleEvent(PublishEvent event) {

        try {

            T typeData = safeExtractType(event.getData());

            queue.push(typeData);

            for (SensorListener<T> listener : listeners) {
                listener.newDataAvailable(typeData);
            }

        } catch (IOException e) {
            logger.error("Error extracting btl type from publisher event data of type " + typeClass.getSimpleName());
            logger.debug(event.getData().getDocumentAsText(), e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getSubscriberName() {
        return subscriberName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean hasNext() {
        return queue.hasNext();
    }

}
