package de.unibi.citec.clf.bonsai.xcf;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import static org.easymock.EasyMock.expect;
import static org.easymock.classextension.EasyMock.*;
import static org.junit.Assert.*;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.sf.xcf.RemoteServer;
import net.sf.xcf.XcfManager;
import net.sf.xcf.naming.NameNotFoundException;
import net.sf.xcf.transport.XOPData;

import org.apache.log4j.BasicConfigurator;
import org.easymock.Capture;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable.ErrorHandlingPolicy;

import Ice.ConnectFailedException;

/**
 * Test cases for {@link RemoteServerReconnectableManager}.
 * 
 * @author jwienke
 */
public class RemoteServerReconnectableManagerTest {

    private XcfManager xcfManagerMock;
    private RemoteServerReconnectableManager manager;

    /**
     * Static initializer for logging.
     */
    @BeforeClass
    public static void setUpLogging() {
        BasicConfigurator.configure();
    }

    /**
     * Set up fixture.
     * 
     * @throws Exception
     *             error...
     */
    @Before
    public void setUp() throws Exception {
        xcfManagerMock = createMock(XcfManager.class);
        manager = new RemoteServerReconnectableManager(xcfManagerMock);
    }

    /**
     * Ensure that a remote server for the desired name is requested at the
     * {@link XcfManager}.
     * 
     * @throws Exception
     *             error
     */
    @Test
    public void initializeRemoteServerCorrect() throws Exception {

        final String remoteServerName = "nameMe";

        RemoteServer remoteServerMock = createMock(RemoteServer.class);

        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                remoteServerMock);

        replay(xcfManagerMock);

        manager.initializeXcfType(remoteServerName);

        verify(xcfManagerMock);

    }

    /**
     * Report error on remote server connection if desired.
     * 
     * @throws Exception
     *             desired error
     */
    @Test(expected = XcfTypeInitializationException.class)
    public void initializeRemoteServerWrongNameError() throws Exception {

        manager.setErrorOnXcfTypeInitialization(true);

        final String remoteServerName = "nameMe";

        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andThrow(
                new NameNotFoundException());

        replay(xcfManagerMock);

        manager.initializeXcfType(remoteServerName);

        verify(xcfManagerMock);

    }

    /**
     * Ignore errors on remote server connection if desired.
     * 
     * @throws Exception
     *             error
     */
    @Test
    public void initializeRemoteServerWrongNameNoError() throws Exception {

        manager.setErrorOnXcfTypeInitialization(false);

        final String remoteServerName = "nameMe";

        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andThrow(
                new NameNotFoundException());

        replay(xcfManagerMock);

        manager.initializeXcfType(remoteServerName);

        verify(xcfManagerMock);

    }

    /**
     * Test that the checking thread is executed.
     * 
     * @throws Exception
     *             error
     */
    @Test
    public void checkingWorks() throws Exception {

        final String remoteServerName = "nameMe";

        RemoteServer remoteServerMock = createMock(RemoteServer.class);

        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                remoteServerMock).times(2, 100);

        replay(xcfManagerMock, remoteServerMock);

        manager = new RemoteServerReconnectableManager(xcfManagerMock, 200);
        manager.initializeXcfType(remoteServerName);

        Thread.sleep(500);

        verify(xcfManagerMock, remoteServerMock);

    }

    /**
     * Tests that the reconnectable receives a remote server.
     * 
     * @throws Exception
     *             error
     */
    @Test
    public void reconnectableGetsRemoteServer() throws Exception {

        final String remoteServerName = "nameMe";

        RemoteServer remoteServerMock = createMock(RemoteServer.class);

        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                remoteServerMock).atLeastOnce();

        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);

        RemoteServerReconnectable reconnectableMock =
                createMock(RemoteServerReconnectable.class);
        expect(reconnectableMock.getRemoteServerNames()).andStubReturn(servers);

        reconnectableMock
                .resetRemoteServer(isA(ConnectionErrorResistantRemoteServer.class));

        replay(xcfManagerMock, remoteServerMock, reconnectableMock);

        manager.initializeAndManage(reconnectableMock);

        verify(xcfManagerMock, remoteServerMock, reconnectableMock);

    }

    /**
     * Tests that the remote server is reset if calling a method on the remote
     * server resulted in an reset of the remote server.
     * 
     * @throws Exception
     *             error
     */
    @Test
    public void remoteServerResetOnErrorFromCall() throws Exception {

        final String remoteServerName = "nameMe";

        RemoteServer firstRemoteServerMock = createMock(RemoteServer.class);
        expect(firstRemoteServerMock.getServerName()).andStubReturn(
                remoteServerName);
        RemoteServer secondRemoteServerMock = createMock(RemoteServer.class);
        expect(secondRemoteServerMock.getServerName()).andStubReturn(
                remoteServerName);

        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                firstRemoteServerMock).times(1);
        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                secondRemoteServerMock).atLeastOnce();

        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);

        RemoteServerReconnectable reconnectableMock =
                createMock(RemoteServerReconnectable.class);
        expect(reconnectableMock.getRemoteServerNames()).andStubReturn(servers);

        Capture<ConnectionErrorResistantRemoteServer> remoteServerCapture =
                new Capture<ConnectionErrorResistantRemoteServer>();
        reconnectableMock.resetRemoteServer(and(
                isA(ConnectionErrorResistantRemoteServer.class),
                capture(remoteServerCapture)));

        final String methodName = "testMethod";
        final XOPData requestData = XOPData.fromString("<request/>");
        final XOPData resultData = XOPData.fromString("<result/>");

        expect(firstRemoteServerMock.callMethod(methodName, requestData))
                .andThrow(new ConnectFailedException());
        expect(secondRemoteServerMock.callMethod(methodName, requestData))
                .andReturn(resultData);

        replay(xcfManagerMock, firstRemoteServerMock, secondRemoteServerMock,
                reconnectableMock);

        final long interval = 100;
        manager =
                new RemoteServerReconnectableManager(xcfManagerMock, interval);
        manager.initializeXcfType(remoteServerName);

        manager.initializeAndManage(reconnectableMock);
        assertTrue(remoteServerCapture.hasCaptured());
        assertEquals(remoteServerCapture.getValue().getWrappedRemoteServer(),
                firstRemoteServerMock);

        XOPData result =
                remoteServerCapture.getValue().callMethod(methodName,
                        requestData);
        assertEquals(resultData, result);

        assertEquals(secondRemoteServerMock, remoteServerCapture.getValue()
                .getWrappedRemoteServer());

        verify(xcfManagerMock, firstRemoteServerMock, secondRemoteServerMock,
                reconnectableMock);

    }

    /**
     * Ensure that the {@link ConnectionErrorResistantRemoteServer} gets a new
     * {@link RemoteServer} even if there was no call that resulted in an error.
     * 
     * @throws Exception
     *             error
     */
    @Test
    public void remoteServerResetWithoutCall() throws Exception {

        final String remoteServerName = "nameMe";

        RemoteServer firstRemoteServerMock = createMock(RemoteServer.class);
        expect(firstRemoteServerMock.getServerName()).andStubReturn(
                remoteServerName);
        RemoteServer secondRemoteServerMock = createMock(RemoteServer.class);
        expect(secondRemoteServerMock.getServerName()).andStubReturn(
                remoteServerName);

        // first have some successful tries for the remote server
        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                firstRemoteServerMock).times(4);
        // then issue an error
        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andThrow(
                new NameNotFoundException()).times(3);
        // and the return a new remote server
        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                secondRemoteServerMock).atLeastOnce();

        // create a mock for the reconnectable
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        RemoteServerReconnectable reconnectableMock =
                createMock(RemoteServerReconnectable.class);
        expect(reconnectableMock.getRemoteServerNames()).andStubReturn(servers);
        // expect that it gets a wrapped remote server
        Capture<ConnectionErrorResistantRemoteServer> remoteServerCapture =
                new Capture<ConnectionErrorResistantRemoteServer>();
        reconnectableMock.resetRemoteServer(and(
                isA(ConnectionErrorResistantRemoteServer.class),
                capture(remoteServerCapture)));

        replay(xcfManagerMock, firstRemoteServerMock, secondRemoteServerMock,
                reconnectableMock);

        final long interval = 100;
        manager =
                new RemoteServerReconnectableManager(xcfManagerMock, interval);
        manager.initializeXcfType(remoteServerName);
        manager.initializeAndManage(reconnectableMock);

        // after initialization there must be the first remote server in the
        // wrapper
        assertEquals(firstRemoteServerMock, remoteServerCapture.getValue()
                .getWrappedRemoteServer());

        Thread.sleep(10 * interval);

        verify(xcfManagerMock, firstRemoteServerMock, secondRemoteServerMock,
                reconnectableMock);

        // at the end we want to have the second remote server in the wrapper
        assertEquals(secondRemoteServerMock, remoteServerCapture.getValue()
                .getWrappedRemoteServer());

    }

    /**
     * Ensure that managing multiple remote servers per
     * {@link RemoteServerReconnectable} works.
     * 
     * @throws Exception
     *             error
     */
    @Test
    public void multipleRemoteServersPerReconnectable() throws Exception {

        final String remoteServerName1 = "server1abc";
        final String remoteServerName2 = "foobar";

        RemoteServer firstRemoteServerMock = createMock(RemoteServer.class);
        expect(firstRemoteServerMock.getServerName()).andStubReturn(
                remoteServerName1);
        RemoteServer secondRemoteServerMock = createMock(RemoteServer.class);
        expect(secondRemoteServerMock.getServerName()).andStubReturn(
                remoteServerName2);

        expect(xcfManagerMock.createRemoteServer(remoteServerName1)).andReturn(
                firstRemoteServerMock);
        expect(xcfManagerMock.createRemoteServer(remoteServerName2)).andReturn(
                secondRemoteServerMock);

        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName1, ErrorHandlingPolicy.SELF_HEALING);
        servers.put(remoteServerName2, ErrorHandlingPolicy.SELF_HEALING);
        RemoteServerReconnectable reconnectableMock =
                createMock(RemoteServerReconnectable.class);
        expect(reconnectableMock.getRemoteServerNames()).andStubReturn(servers);

        Capture<ConnectionErrorResistantRemoteServer> remoteServerCapture1 =
                new Capture<ConnectionErrorResistantRemoteServer>();
        reconnectableMock.resetRemoteServer(and(
                isA(ConnectionErrorResistantRemoteServer.class),
                capture(remoteServerCapture1)));
        Capture<ConnectionErrorResistantRemoteServer> remoteServerCapture2 =
                new Capture<ConnectionErrorResistantRemoteServer>();
        reconnectableMock.resetRemoteServer(and(
                isA(ConnectionErrorResistantRemoteServer.class),
                capture(remoteServerCapture2)));

        replay(xcfManagerMock, firstRemoteServerMock, secondRemoteServerMock,
                reconnectableMock);

        manager.initializeXcfType(remoteServerName1);
        manager.initializeXcfType(remoteServerName2);
        manager.initializeAndManage(reconnectableMock);

        verify(xcfManagerMock, firstRemoteServerMock, secondRemoteServerMock,
                reconnectableMock);

        assertTrue(remoteServerCapture1.hasCaptured());
        assertTrue(remoteServerCapture2.hasCaptured());
        boolean containsFirst = false;
        boolean containsSecond = false;
        List<ConnectionErrorResistantRemoteServer> capturedServers =
                new LinkedList<ConnectionErrorResistantRemoteServer>();
        capturedServers.add(remoteServerCapture1.getValue());
        capturedServers.add(remoteServerCapture2.getValue());
        for (ConnectionErrorResistantRemoteServer server : capturedServers) {
            if (server.getWrappedRemoteServer().equals(firstRemoteServerMock)) {
                containsFirst = true;
            }
            if (server.getWrappedRemoteServer().equals(secondRemoteServerMock)) {
                containsSecond = true;
            }
        }
        assertTrue(containsFirst);
        assertTrue(containsSecond);

    }

    /**
     * Ensures that the wish to have a direct notification for a new remote
     * server after a connection loss is obeyed.
     * 
     * @throws Exception
     *             error
     */
    @Test
    public void directNotificationReset() throws Exception {

        final String remoteServerName = "nameMe";

        RemoteServer firstRemoteServerMock = createMock(RemoteServer.class);
        expect(firstRemoteServerMock.getServerName()).andStubReturn(
                remoteServerName);
        RemoteServer secondRemoteServerMock = createMock(RemoteServer.class);
        expect(secondRemoteServerMock.getServerName()).andStubReturn(
                remoteServerName);

        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                firstRemoteServerMock).times(1);
        expect(xcfManagerMock.createRemoteServer(remoteServerName)).andReturn(
                secondRemoteServerMock).atLeastOnce();

        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.NOTIFYING);

        RemoteServerReconnectable reconnectableMock =
                createMock(RemoteServerReconnectable.class);
        expect(reconnectableMock.getRemoteServerNames()).andStubReturn(servers);

        Capture<ConnectionErrorNotifyingRemoteServer> firstRemoteServerCapture =
                new Capture<ConnectionErrorNotifyingRemoteServer>();
        reconnectableMock.resetRemoteServer(and(
                isA(ConnectionErrorNotifyingRemoteServer.class),
                capture(firstRemoteServerCapture)));
        expectLastCall().once();
        Capture<ConnectionErrorNotifyingRemoteServer> secondRemoteServerCapture =
                new Capture<ConnectionErrorNotifyingRemoteServer>();
        reconnectableMock.resetRemoteServer(and(
                isA(ConnectionErrorNotifyingRemoteServer.class),
                capture(secondRemoteServerCapture)));
        expectLastCall().atLeastOnce();

        final String methodName = "testMethod";
        final XOPData requestData = XOPData.fromString("<request/>");
        final XOPData resultData = XOPData.fromString("<result/>");

        expect(firstRemoteServerMock.callMethod(methodName, requestData))
                .andThrow(new ConnectFailedException());
        expect(secondRemoteServerMock.callMethod(methodName, requestData))
                .andReturn(resultData).atLeastOnce();

        replay(xcfManagerMock, firstRemoteServerMock, secondRemoteServerMock,
                reconnectableMock);

        final long interval = 100;
        manager =
                new RemoteServerReconnectableManager(xcfManagerMock, interval);
        manager.initializeXcfType(remoteServerName);

        manager.initializeAndManage(reconnectableMock);

        try {
            assertTrue(firstRemoteServerCapture.hasCaptured());
            RemoteServer capturedServer = firstRemoteServerCapture.getValue();
            capturedServer.callMethod(methodName, requestData);
            fail("first call to this must fail");
            // CHECKSTYLE:OFF
        } catch (Exception e) {
            // CHECKSTYLE:ON
            // this is wanted
        }

        // CHECKSTYLE:OFF after this time there must be a new remote server set
        Thread.sleep(4 * interval);
        // CHECKSTYLE:ON

        assertTrue(secondRemoteServerCapture.hasCaptured());
        RemoteServer capturedServer = secondRemoteServerCapture.getValue();
        XOPData result = capturedServer.callMethod(methodName, requestData);
        assertEquals(resultData, result);

        verify(xcfManagerMock, firstRemoteServerMock, secondRemoteServerMock,
                reconnectableMock);

        System.out.println(System.identityHashCode(firstRemoteServerCapture
                .getValues().get(0)));

        assertEquals(firstRemoteServerCapture.getValue()
                .getWrappedRemoteServer(), firstRemoteServerMock);
        assertEquals(secondRemoteServerCapture.getValue()
                .getWrappedRemoteServer(), secondRemoteServerMock);

    }

}
